from __future__ import annotations

import logging
from numbers import Integral
from typing import TYPE_CHECKING
from urllib.error import HTTPError
from urllib.parse import unquote, urlparse
from urllib.request import Request, urlopen

import deltalake
from deltalake.exceptions import TableNotFoundError
from deltalake.table import DeltaTable
from packaging.version import parse

import daft
import daft.exceptions
from daft.daft import (
    FileFormatConfig,
    ParquetSourceConfig,
    PyPartitionField,
    PyPushdowns,
    S3Config,
    ScanTask,
    StorageConfig,
)
from daft.io.delta_lake.utils import construct_delta_file_path
from daft.io.object_store_options import io_config_to_storage_options
from daft.io.scan import ScanOperator
from daft.logical.schema import Schema

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

logger = logging.getLogger(__name__)


def get_s3_bucket_region(bucket_name: str) -> str | None:
    # When making a https request to https://{bucket_name}.s3.amazonaws.com/, aws returns either a 200 response, 403 response, or 404 response.
    # In the header of the 200 and 403 responses, there is an `x-amz-bucket-region` field from which we can extract the bucket's region.
    url = f"https://{bucket_name}.s3.amazonaws.com"

    try:
        req = Request(url, method="HEAD")
        with urlopen(req) as response:
            return response.headers.get("x-amz-bucket-region")
    except HTTPError as e:
        bucket_region = e.headers.get("x-amz-bucket-region")
        if bucket_region is None:
            logger.warning(
                "Failed to get the S3 bucket region using the given S3 uri. HTTPError error: %s",
                e,
            )
        return bucket_region
    except Exception as e:
        logger.warning(
            "Failed to get the S3 bucket region using the given S3 uri. Error: %s",
            e,
        )
        return None


class DeltaLakeScanOperator(ScanOperator):
    def __init__(
        self,
        table_uri: str,
        storage_config: StorageConfig,
        version: int | str | datetime | None = None,
        ignore_deletion_vectors: bool = False,
    ) -> None:
        super().__init__()

        # Unfortunately delta-rs doesn't do very good inference of credentials for S3. Thus the current Daft behavior of passing
        # in `None` for credentials will cause issues when instantiating the DeltaTable without credentials.
        #
        # Thus, if we don't detect any credentials being available, we attempt to detect it from the environment using our Daft credentials chain.
        #
        # See: https://github.com/delta-io/delta-rs/issues/2117
        deltalake_sdk_io_config = storage_config.io_config
        scheme = urlparse(table_uri).scheme
        if scheme == "s3" or scheme == "s3a":
            # Try to get the bucket's region.
            if deltalake_sdk_io_config.s3.region_name is None:
                bucket_name = urlparse(table_uri).netloc
                region = get_s3_bucket_region(bucket_name)
                if region is not None:
                    deltalake_sdk_io_config = deltalake_sdk_io_config.replace(
                        s3=deltalake_sdk_io_config.s3.replace(region_name=region)
                    )

            # Try to get config from the environment
            if any(
                [
                    deltalake_sdk_io_config.s3.key_id is None,
                    deltalake_sdk_io_config.s3.region_name is None,
                ]
            ):
                try:
                    s3_config_from_env = S3Config.from_env()
                # Sometimes S3Config.from_env throws an error, for example on CI machines with weird metadata servers.
                except daft.exceptions.DaftCoreException:
                    pass
                else:
                    if (
                        deltalake_sdk_io_config.s3.key_id is None
                        and deltalake_sdk_io_config.s3.access_key is None
                        and deltalake_sdk_io_config.s3.session_token is None
                    ):
                        deltalake_sdk_io_config = deltalake_sdk_io_config.replace(
                            s3=deltalake_sdk_io_config.s3.replace(
                                key_id=s3_config_from_env.key_id,
                                access_key=s3_config_from_env.access_key,
                                session_token=s3_config_from_env.session_token,
                            )
                        )
                    if deltalake_sdk_io_config.s3.region_name is None:
                        deltalake_sdk_io_config = deltalake_sdk_io_config.replace(
                            s3=deltalake_sdk_io_config.s3.replace(
                                region_name=s3_config_from_env.region_name,
                            )
                        )
        elif scheme == "gcs" or scheme == "gs":
            # TO-DO: Handle any key-value replacements in `io_config` if there are missing elements
            pass
        elif scheme == "az" or scheme == "abfs" or scheme == "abfss":
            # TO-DO: Handle any key-value replacements in `io_config` if there are missing elements
            pass

        self._table = DeltaTable(
            table_uri,
            storage_options=io_config_to_storage_options(deltalake_sdk_io_config, table_uri),
        )

        if version is not None:
            if isinstance(version, Integral) and not isinstance(version, bool):
                requested_version = int(version)
                latest_version = self._table.version()
                if requested_version < 0 or requested_version > latest_version:
                    raise TableNotFoundError(
                        f"Requested Delta Lake table version {requested_version} does not exist. "
                        f"Latest available version is {latest_version}."
                    )
                self._table.load_as_version(requested_version)
            else:
                self._table.load_as_version(version)

        self._storage_config = storage_config

        from ._deltalake import delta_schema_to_pyarrow

        self._schema = Schema.from_pyarrow_schema(delta_schema_to_pyarrow(self._table.schema()))
        partition_columns = set(self._table.metadata().partition_columns)
        self._partition_keys = [
            PyPartitionField(field._field) for field in self._schema if field.name in partition_columns
        ]
        self._ignore_deletion_vectors = ignore_deletion_vectors

    def schema(self) -> Schema:
        return self._schema

    def name(self) -> str:
        return "DeltaLakeScanOperator"

    def display_name(self) -> str:
        return f"DeltaLakeScanOperator({self._table.metadata().name})"

    def partitioning_keys(self) -> list[PyPartitionField]:
        return self._partition_keys

    def multiline_display(self) -> list[str]:
        return [
            self.display_name(),
            f"Schema = {self._schema}",
            f"Partitioning keys = {self.partitioning_keys()}",
            # TODO(Clark): Improve repr of storage config here.
            f"Storage config = {self._storage_config}",
        ]

    def to_scan_tasks(self, pushdowns: PyPushdowns) -> Iterator[ScanTask]:
        import pyarrow as pa

        metadata = self._table.metadata()
        deletion_vectors_enabled = metadata.configuration.get("delta.enableDeletionVectors", "false").lower() == "true"
        # If deletion vectors are enabled and the deltalake library does not support propagation
        # raise an error if ignore_deletion_vectors is not set.
        if deletion_vectors_enabled and _missing_deletion_vectors_propagation():
            if not self._ignore_deletion_vectors:
                raise NotImplementedError(
                    "Delta Lake deletion vectors are not yet supported; please let the Daft team know if you'd like to see this feature!\n"
                    "Note that for Delta Lake versions 1.2.0 and newer, deletion vectors are not included in add actions.\n"
                    "Deletion records can be dropped from this table to allow it to be read with Daft: https://docs.delta.io/latest/delta-drop-feature.html\n"
                    "Alternatively, you can set ignore_deletion_vectors=True to skip checking for deletion vectors."
                )

        # TODO(Clark): Push limit and filter expressions into deltalake action fetch, to prune the files returned.
        # Issue: https://github.com/Eventual-Inc/Daft/issues/1953
        add_actions = pa.record_batch(self._table.get_add_actions())
        if len(self.partitioning_keys()) > 0 and pushdowns.partition_filters is None:
            logger.warning(
                "%s has partitioning keys = %s, but no partition filter was specified. This will result in a full table scan.",
                self.display_name(),
                self.partitioning_keys(),
            )

        # TODO(Clark): Add support for deletion vectors.
        # Issue: https://github.com/Eventual-Inc/Daft/issues/1954
        if not self._ignore_deletion_vectors and "deletionVector" in add_actions.schema.names:
            raise NotImplementedError(
                "Delta Lake deletion vectors are not yet supported; please let the Daft team know if you'd like to see this feature!\n"
                "Deletion records can be dropped from this table to allow it to be read with Daft: https://docs.delta.io/latest/delta-drop-feature.html\n"
                "Alternatively, you can set ignore_deletion_vectors=True to skip checking for deletion vectors."
            )

        # TODO(Clark): Add support for column mappings.
        # Issue: https://github.com/Eventual-Inc/Daft/issues/1955

        limit_files = pushdowns.limit is not None and pushdowns.filters is None and pushdowns.partition_filters is None
        rows_left = pushdowns.limit if pushdowns.limit is not None else 0
        scan_tasks = []

        # Determine which partition field name is used in the schema
        # Deltalake versions <1.2.0 use "partition_values", >=1.2.0 use "partition"
        is_deltalake_v1_2_or_above = parse(deltalake.__version__) >= parse("1.2.0")

        partition_field_name = "partition" if is_deltalake_v1_2_or_above else "partition_values"
        is_partitioned = (
            partition_field_name in add_actions.schema.names
            and add_actions.schema.field(partition_field_name).type.num_fields > 0
        )
        for task_idx in range(add_actions.num_rows):
            if limit_files and rows_left <= 0:
                break

            # NOTE: The paths in the transaction log consist of the post-table-uri suffix.
            # Workaround for deltalake >= 1.2.0: paths are double-encoded in the log but single-encoded on disk.
            scheme = urlparse(self._table.table_uri).scheme
            raw_path = add_actions["path"][task_idx].as_py()
            if is_deltalake_v1_2_or_above:
                # For Delta Lake >= 1.2.0, decode double-encoded paths
                unquoted = unquote(raw_path)
                decoded_path = unquoted if unquoted != raw_path else raw_path
            else:
                # For Delta Lake < 1.2.0, use path as-is
                decoded_path = raw_path
            path = construct_delta_file_path(scheme, self._table.table_uri, decoded_path)

            try:
                record_count = add_actions["num_records"][task_idx].as_py()
            except KeyError:
                record_count = None

            try:
                size_bytes = add_actions["size_bytes"][task_idx].as_py()
            except KeyError:
                size_bytes = None
            file_format_config = FileFormatConfig.from_parquet_config(ParquetSourceConfig())

            if is_partitioned:
                dtype = add_actions.schema.field(partition_field_name).type
                part_values = add_actions[partition_field_name][task_idx]
                arrays = {}
                for field_idx in range(dtype.num_fields):
                    field_name = dtype.field(field_idx).name
                    try:
                        arrow_arr = pa.array([part_values[field_name]], type=dtype.field(field_idx).type)
                    except (
                        pa.ArrowInvalid,
                        pa.ArrowTypeError,
                        pa.ArrowNotImplementedError,
                    ):
                        # pyarrow < 13.0.0 doesn't accept pyarrow scalars in the array constructor.
                        arrow_arr = pa.array(
                            [part_values[field_name].as_py()],
                            type=dtype.field(field_idx).type,
                        )
                    arrays[field_name] = daft.Series.from_arrow(arrow_arr, field_name)
                partition_values = daft.recordbatch.RecordBatch.from_pydict(arrays)._recordbatch
            else:
                partition_values = None

            # Populate scan task with column-wise stats.
            schema_names = add_actions.schema.names
            if "min" in schema_names and "max" in schema_names:
                dtype = add_actions.schema.field("min").type
                min_values = add_actions["min"][task_idx]
                max_values = add_actions["max"][task_idx]
                # TODO(Clark): Add support for tracking null counts in column stats.
                # null_counts = add_actions["null_count"][task_idx]
                arrays = {}
                for field_idx in range(dtype.num_fields):
                    field_name = dtype.field(field_idx).name
                    try:
                        arrow_arr = pa.array(
                            [min_values[field_name], max_values[field_name]],
                            type=dtype.field(field_idx).type,
                        )
                    except (
                        pa.ArrowInvalid,
                        pa.ArrowTypeError,
                        pa.ArrowNotImplementedError,
                    ):
                        # pyarrow < 13.0.0 doesn't accept pyarrow scalars in the array constructor.
                        arrow_arr = pa.array(
                            [
                                min_values[field_name].as_py(),
                                max_values[field_name].as_py(),
                            ],
                            type=dtype.field(field_idx).type,
                        )
                    arrays[field_name] = daft.Series.from_arrow(arrow_arr, field_name)
                stats = daft.recordbatch.RecordBatch.from_pydict(arrays)
            else:
                stats = None
            st = ScanTask.catalog_scan_task(
                file=path,
                file_format=file_format_config,
                schema=self._schema._schema,
                num_rows=record_count,
                storage_config=self._storage_config,
                size_bytes=size_bytes,
                iceberg_delete_files=None,
                pushdowns=pushdowns,
                partition_values=partition_values,
                stats=stats._recordbatch if stats is not None else None,
            )
            if st is None:
                continue
            if record_count is not None:
                rows_left -= record_count
            scan_tasks.append(st)
        return iter(scan_tasks)

    def can_absorb_filter(self) -> bool:
        return False

    def can_absorb_limit(self) -> bool:
        return False

    def can_absorb_select(self) -> bool:
        return True


# Returns True if the Delta Lake library does not return the deletion vectors in the add_actions.
# These were removed in deltalake>=1.2.0.
def _missing_deletion_vectors_propagation() -> bool:
    return parse(deltalake.__version__) >= parse("1.2.0")



from __future__ import annotations

import pathlib
from itertools import chain
from typing import TYPE_CHECKING, Any, Literal

from daft.context import get_context
from daft.datatype import DataType
from daft.dependencies import pa
from daft.io import DataSink
from daft.io.sink import WriteResult
from daft.recordbatch import MicroPartition
from daft.schema import Schema

# Alias for Lance fragment metadata type to avoid depending on Lance type hints at type-check time
LanceFragmentMetadata = Any

if TYPE_CHECKING:
    from collections.abc import Iterator
    from types import ModuleType

    from daft.daft import IOConfig


def pyarrow_schema_castable(src: pa.Schema, dst: pa.Schema) -> bool:
    if len(src) != len(dst):
        return False
    for src_field, dst_field in zip(src, dst):
        empty_array = pa.array([], type=src_field.type)
        try:
            empty_array.cast(dst_field.type)
        except Exception:
            return False
    return True


def table_schema_subset_castable(data_schema: pa.Schema, table_schema: pa.Schema) -> bool:
    """Checks whether data_schema is a superset of table_schema and every table field is castable from data.

    For each field in the existing table schema, ensure there is a field with the same name in the data schema,
    and that Arrow can cast from the data field type to the table field type.
    """
    data_fields_by_name = {f.name: f for f in data_schema}
    for table_field in table_schema:
        src_field = data_fields_by_name.get(table_field.name)
        if src_field is None:
            return False
        empty_array = pa.array([], type=src_field.type)
        try:
            empty_array.cast(table_field.type)
        except Exception:
            return False
    return True


class LanceDataSink(DataSink[list[LanceFragmentMetadata]]):
    """WriteSink for writing data to a Lance dataset."""

    def _import_lance(self) -> ModuleType:
        try:
            import lance

            return lance
        except ImportError:
            raise ImportError("lance is not installed. Please install lance using `pip install daft[lance]`")

    def __init__(
        self,
        uri: str | pathlib.Path,
        schema: Schema | pa.Schema,
        mode: Literal["create", "append", "overwrite", "merge"],
        io_config: IOConfig | None = None,
        **kwargs: Any,
    ) -> None:
        from daft.io.object_store_options import io_config_to_storage_options

        lance = self._import_lance()
        if not isinstance(uri, (str, pathlib.Path)):
            raise TypeError(f"Expected URI to be str or pathlib.Path, got {type(uri)}")
        self._table_uri = str(uri)
        self._mode = mode
        self._io_config = get_context().daft_planning_config.default_io_config if io_config is None else io_config
        self._kwargs = kwargs
        # Strip merge-specific kwargs that are not applicable to non-merge write paths
        self._kwargs.pop("left_on", None)
        self._kwargs.pop("right_on", None)
        self._storage_options = io_config_to_storage_options(self._io_config, self._table_uri)

        if isinstance(schema, Schema):
            self._pyarrow_schema = schema.to_pyarrow_schema()
        elif isinstance(schema, pa.Schema):
            self._pyarrow_schema = schema
        else:
            raise TypeError(f"Expected schema to be Schema or pa.Schema, got {type(schema)}")

        # Filter out internal metadata fields that should not be part of the validation
        # These fields are used internally by Lance but should not affect schema validation
        meta_exclusions = {"fragment_id", "_rowaddr", "_rowid"}
        # Create a filtered schema without internal metadata fields for validation
        filtered_fields = [field for field in self._pyarrow_schema if field.name not in meta_exclusions]
        self._filtered_schema = pa.schema(filtered_fields)

        self._version: int = 0
        self._table_schema: pa.Schema | None = None

        table = self._load_existing_dataset(lance)
        self._validate_dataset(table)
        self._schema = Schema._from_field_name_and_types(
            [
                ("num_fragments", DataType.int64()),
                ("num_deleted_rows", DataType.int64()),
                ("num_small_files", DataType.int64()),
                ("version", DataType.int64()),
            ]
        )

    def _load_existing_dataset(self, lance_module: ModuleType) -> Any | None:
        try:
            return lance_module.dataset(self._table_uri, storage_options=self._storage_options)
        except (ValueError, FileNotFoundError, OSError) as e:
            # Check if this is specifically a "dataset not found" error
            if "not found" in str(e).lower() or "does not exist" in str(e).lower():
                if self._mode == "append":
                    raise ValueError("Cannot append to non-existent Lance dataset.")
                return None
            else:
                # Re-raise other errors (permissions, network, etc.)
                raise

    def _validate_dataset(self, table: Any) -> None:
        """Validate dataset against the current mode."""
        if table is None:
            if self._mode == "append" or self._mode == "merge":
                raise ValueError("Cannot append or merge to non-existent Lance dataset.")
            if self._mode == "create" and self._storage_options is None:
                p = pathlib.Path(self._table_uri)
                if p.is_file():
                    raise FileExistsError("Target path points to a file, cannot create a dataset here.")
        else:
            self._table_schema = table.schema
            self._version = table.latest_version

            if self._mode == "create":
                raise ValueError("Cannot create a Lance dataset at a location where one already exists.")

            if self._mode == "append" and not pyarrow_schema_castable(self._pyarrow_schema, self._table_schema):
                raise ValueError(
                    "Schema of data does not match table schema\n"
                    f"Data schema:\n{self._pyarrow_schema}\nTable Schema:\n{self._table_schema}"
                )

    def name(self) -> str:
        """Optional custom sink name."""
        return "Lance Write"

    def schema(self) -> Schema:
        return self._schema

    def write(self, micropartitions: Iterator[MicroPartition]) -> Iterator[WriteResult[list[LanceFragmentMetadata]]]:
        """Writes fragments from the given micropartitions."""
        lance = self._import_lance()

        max_rows_per_file = self._kwargs.get("max_rows_per_file")
        accumulate = isinstance(max_rows_per_file, int) and max_rows_per_file > 0
        threshold: int = max_rows_per_file if isinstance(max_rows_per_file, int) else 0

        def _prepare_arrow_table(input_table: pa.Table) -> pa.Table:
            if self._mode == "merge":
                target_schema = self._filtered_schema
            else:
                target_schema = self._table_schema
            if target_schema is None:
                target_schema = self._pyarrow_schema
            if self._table_schema is not None:
                return input_table.cast(self._table_schema)
            if not pa.Schema.equals(target_schema, input_table.schema):
                return input_table.cast(target_schema)
            return pa.Table.from_batches(input_table.to_batches(), target_schema)

        def _write_arrow_table(table: pa.Table) -> WriteResult[list[LanceFragmentMetadata]]:
            fragments = lance.fragment.write_fragments(
                table,
                dataset_uri=self._table_uri,
                mode=self._mode,
                storage_options=self._storage_options,
                **self._kwargs,
            )
            return WriteResult(result=fragments, bytes_written=table.nbytes, rows_written=table.num_rows)

        class _TableAccumulator:
            def __init__(self, threshold: int):
                self._threshold = threshold
                self._batches: list[pa.RecordBatch] = []
                self._rows = 0
                self._schema: pa.Schema | None = None

            def add(self, table: pa.Table) -> bool:
                if self._schema is None:
                    self._schema = table.schema
                for b in table.to_batches():
                    self._batches.append(b)
                self._rows += table.num_rows
                return self._rows >= self._threshold

            def has_rows(self) -> bool:
                return self._rows > 0

            def build_and_reset(self) -> pa.Table:
                table = pa.Table.from_batches(self._batches, schema=self._schema)
                self._batches = []
                self._rows = 0
                return table

        accumulator = _TableAccumulator(threshold) if accumulate else None

        for micropartition in micropartitions:
            input_table = micropartition.to_arrow()
            arrow_table = _prepare_arrow_table(input_table)

            if not accumulate:
                yield _write_arrow_table(arrow_table)
                continue

            # If a single micropartition exceeds threshold, flush accumulated first then write directly
            if arrow_table.num_rows >= threshold:
                if accumulator and accumulator.has_rows():
                    yield _write_arrow_table(accumulator.build_and_reset())
                yield _write_arrow_table(arrow_table)
                continue

            # Otherwise, accumulate and flush when threshold reached
            if accumulator and accumulator.add(arrow_table):
                yield _write_arrow_table(accumulator.build_and_reset())

        # Flush any remaining accumulated rows
        if accumulate and accumulator and accumulator.has_rows():
            yield _write_arrow_table(accumulator.build_and_reset())

    def finalize(self, write_results: list[WriteResult[list[LanceFragmentMetadata]]]) -> MicroPartition:
        """Commits the fragments to the Lance dataset. Returns a DataFrame with the stats of the dataset."""
        lance = self._import_lance()

        fragments = list(chain.from_iterable(write_result.result for write_result in write_results))

        if self._mode == "create" or self._mode == "overwrite":
            operation = lance.LanceOperation.Overwrite(self._pyarrow_schema, fragments)
        elif self._mode == "append":
            operation = lance.LanceOperation.Append(fragments)

        dataset = lance.LanceDataset.commit(
            self._table_uri,
            operation,
            read_version=self._version,
            storage_options=self._storage_options,
        )
        stats = dataset.stats.dataset_stats()

        tbl = MicroPartition.from_pydict(
            {
                "num_fragments": pa.array([stats["num_fragments"]], type=pa.int64()),
                "num_deleted_rows": pa.array([stats["num_deleted_rows"]], type=pa.int64()),
                "num_small_files": pa.array([stats["num_small_files"]], type=pa.int64()),
                "version": pa.array([dataset.version], type=pa.int64()),
            }
        )
        return tbl



"""Date and Time Functions."""

from __future__ import annotations

from typing import TYPE_CHECKING

from daft.expressions import Expression

if TYPE_CHECKING:
    from daft.datatype import TimeUnit


def date(expr: Expression) -> Expression:
    """Retrieves the date for a datetime column.

    Returns:
        Expression: a Date expression

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import date
        >>> df = daft.from_pydict(
        ...     {
        ...         "x": [
        ...             datetime.datetime(2021, 1, 1, 5, 1, 1),
        ...             datetime.datetime(2021, 1, 2, 6, 1, 59),
        ...             datetime.datetime(2021, 1, 3, 7, 2, 0),
        ...         ],
        ...     }
        ... )
        >>> df = df.with_column("date", date(df["x"]))
        >>> df.show()
        ╭─────────────────────┬────────────╮
        │ x                   ┆ date       │
        │ ---                 ┆ ---        │
        │ Timestamp[us]       ┆ Date       │
        ╞═════════════════════╪════════════╡
        │ 2021-01-01 05:01:01 ┆ 2021-01-01 │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2021-01-02 06:01:59 ┆ 2021-01-02 │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2021-01-03 07:02:00 ┆ 2021-01-03 │
        ╰─────────────────────┴────────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("date", expr)


def day(expr: Expression) -> Expression:
    """Retrieves the day for a datetime column.

    Returns:
        Expression: a UInt32 expression with just the day extracted from a datetime column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import day
        >>> df = daft.from_pydict(
        ...     {
        ...         "x": [
        ...             datetime.datetime(2021, 1, 1, 5, 1, 1),
        ...             datetime.datetime(2021, 1, 2, 6, 1, 59),
        ...             datetime.datetime(2021, 1, 3, 7, 2, 0),
        ...         ],
        ...     }
        ... )
        >>> df = df.with_column("day", day(df["x"]))
        >>> df.show()
        ╭─────────────────────┬────────╮
        │ x                   ┆ day    │
        │ ---                 ┆ ---    │
        │ Timestamp[us]       ┆ UInt32 │
        ╞═════════════════════╪════════╡
        │ 2021-01-01 05:01:01 ┆ 1      │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ 2021-01-02 06:01:59 ┆ 2      │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ 2021-01-03 07:02:00 ┆ 3      │
        ╰─────────────────────┴────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("day", expr)


def hour(expr: Expression) -> Expression:
    """Retrieves the hour for a datetime column.

    Returns:
        Expression: a UInt32 expression with just the hour extracted from a datetime column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import hour
        >>> df = daft.from_pydict(
        ...     {
        ...         "x": [
        ...             datetime.datetime(2021, 1, 1, 5, 1, 1),
        ...             datetime.datetime(2021, 1, 2, 6, 1, 59),
        ...             datetime.datetime(2021, 1, 3, 7, 2, 0),
        ...         ],
        ...     }
        ... )
        >>> df = df.with_column("hour", hour(df["x"]))
        >>> df.show()
        ╭─────────────────────┬────────╮
        │ x                   ┆ hour   │
        │ ---                 ┆ ---    │
        │ Timestamp[us]       ┆ UInt32 │
        ╞═════════════════════╪════════╡
        │ 2021-01-01 05:01:01 ┆ 5      │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ 2021-01-02 06:01:59 ┆ 6      │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ 2021-01-03 07:02:00 ┆ 7      │
        ╰─────────────────────┴────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("hour", expr)


def minute(expr: Expression) -> Expression:
    """Retrieves the minute for a datetime column.

    Returns:
        Expression: a UInt32 expression with just the minute extracted from a datetime column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import minute
        >>> df = daft.from_pydict(
        ...     {
        ...         "x": [
        ...             datetime.datetime(2021, 1, 1, 5, 1, 1),
        ...             datetime.datetime(2021, 1, 2, 6, 1, 59),
        ...             datetime.datetime(2021, 1, 3, 7, 2, 0),
        ...         ],
        ...     }
        ... )
        >>> df = df.with_column("minute", minute(df["x"]))
        >>> df.show()
        ╭─────────────────────┬────────╮
        │ x                   ┆ minute │
        │ ---                 ┆ ---    │
        │ Timestamp[us]       ┆ UInt32 │
        ╞═════════════════════╪════════╡
        │ 2021-01-01 05:01:01 ┆ 1      │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ 2021-01-02 06:01:59 ┆ 1      │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ 2021-01-03 07:02:00 ┆ 2      │
        ╰─────────────────────┴────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("minute", expr)


def second(expr: Expression) -> Expression:
    """Retrieves the second for a datetime column.

    Returns:
        Expression: a UInt32 expression with just the second extracted from a datetime column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import second
        >>> df = daft.from_pydict(
        ...     {
        ...         "x": [
        ...             datetime.datetime(2021, 1, 1, 0, 1, 1),
        ...             datetime.datetime(2021, 1, 1, 0, 1, 59),
        ...             datetime.datetime(2021, 1, 1, 0, 2, 0),
        ...         ],
        ...     }
        ... )
        >>> df = df.with_column("second", second(df["x"]))
        >>> df.show()
        ╭─────────────────────┬────────╮
        │ x                   ┆ second │
        │ ---                 ┆ ---    │
        │ Timestamp[us]       ┆ UInt32 │
        ╞═════════════════════╪════════╡
        │ 2021-01-01 00:01:01 ┆ 1      │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ 2021-01-01 00:01:59 ┆ 59     │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ 2021-01-01 00:02:00 ┆ 0      │
        ╰─────────────────────┴────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("second", expr)


def millisecond(expr: Expression) -> Expression:
    """Retrieves the millisecond for a datetime column.

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import millisecond
        >>> df = daft.from_pydict(
        ...     {
        ...         "datetime": [
        ...             datetime.datetime(1978, 1, 1, 1, 1, 1, 0),
        ...             datetime.datetime(2024, 10, 13, 5, 30, 14, 500_000),
        ...             datetime.datetime(2065, 1, 1, 10, 20, 30, 60_000),
        ...         ]
        ...     }
        ... )
        >>> df = df.select(millisecond(df["datetime"]))
        >>> df.show()
        ╭──────────╮
        │ datetime │
        │ ---      │
        │ UInt32   │
        ╞══════════╡
        │ 0        │
        ├╌╌╌╌╌╌╌╌╌╌┤
        │ 500      │
        ├╌╌╌╌╌╌╌╌╌╌┤
        │ 60       │
        ╰──────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    return Expression._call_builtin_scalar_fn("millisecond", expr)


def microsecond(expr: Expression) -> Expression:
    """Retrieves the microsecond for a datetime column.

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import microsecond
        >>> df = daft.from_pydict(
        ...     {
        ...         "datetime": [
        ...             datetime.datetime(1978, 1, 1, 1, 1, 1, 0),
        ...             datetime.datetime(2024, 10, 13, 5, 30, 14, 500_000),
        ...             datetime.datetime(2065, 1, 1, 10, 20, 30, 60_000),
        ...         ]
        ...     }
        ... )
        >>> df.select(microsecond(df["datetime"])).show()
        ╭──────────╮
        │ datetime │
        │ ---      │
        │ UInt32   │
        ╞══════════╡
        │ 0        │
        ├╌╌╌╌╌╌╌╌╌╌┤
        │ 500000   │
        ├╌╌╌╌╌╌╌╌╌╌┤
        │ 60000    │
        ╰──────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("microsecond", expr)


def nanosecond(expr: Expression) -> Expression:
    """Retrieves the nanosecond for a datetime column.

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import nanosecond
        >>> df = daft.from_pydict(
        ...     {
        ...         "datetime": [
        ...             datetime.datetime(1978, 1, 1, 1, 1, 1, 0),
        ...             datetime.datetime(2024, 10, 13, 5, 30, 14, 500_000),
        ...             datetime.datetime(2065, 1, 1, 10, 20, 30, 60_000),
        ...         ]
        ...     }
        ... )
        >>>
        >>> df.select(nanosecond(df["datetime"])).show()
        ╭───────────╮
        │ datetime  │
        │ ---       │
        │ UInt32    │
        ╞═══════════╡
        │ 0         │
        ├╌╌╌╌╌╌╌╌╌╌╌┤
        │ 500000000 │
        ├╌╌╌╌╌╌╌╌╌╌╌┤
        │ 60000000  │
        ╰───────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("nanosecond", expr)


def unix_date(expr: Expression) -> Expression:
    """Retrieves the number of days since 1970-01-01 00:00:00 UTC.

    Returns:
        Expression: a UInt64 expression

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import unix_date
        >>> df = daft.from_pydict(
        ...     {
        ...         "datetime": [
        ...             datetime.datetime(1978, 1, 1, 1, 1, 1, 0),
        ...             datetime.datetime(2024, 10, 13, 5, 30, 14, 500_000),
        ...             datetime.datetime(2065, 1, 1, 10, 20, 30, 60_000),
        ...         ]
        ...     }
        ... )
        >>>
        >>> df.select(unix_date(df["datetime"]).alias("unix_date")).show()
        ╭───────────╮
        │ unix_date │
        │ ---       │
        │ UInt64    │
        ╞═══════════╡
        │ 2922      │
        ├╌╌╌╌╌╌╌╌╌╌╌┤
        │ 20009     │
        ├╌╌╌╌╌╌╌╌╌╌╌┤
        │ 34699     │
        ╰───────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("unix_date", expr)


def time(expr: Expression) -> Expression:
    """Retrieves the time for a datetime column.

    Returns:
        Expression: a Time expression

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import time
        >>> df = daft.from_pydict(
        ...     {
        ...         "x": [
        ...             datetime.datetime(2021, 1, 1, 0, 1, 1),
        ...             datetime.datetime(2021, 1, 1, 12, 1, 59),
        ...             datetime.datetime(2021, 1, 1, 23, 59, 59),
        ...         ],
        ...     }
        ... )
        >>> df = df.with_column("time", time(df["x"]))
        >>> df.show()
        ╭─────────────────────┬──────────╮
        │ x                   ┆ time     │
        │ ---                 ┆ ---      │
        │ Timestamp[us]       ┆ Time[us] │
        ╞═════════════════════╪══════════╡
        │ 2021-01-01 00:01:01 ┆ 00:01:01 │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌┤
        │ 2021-01-01 12:01:59 ┆ 12:01:59 │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌┤
        │ 2021-01-01 23:59:59 ┆ 23:59:59 │
        ╰─────────────────────┴──────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("time", expr)


def month(expr: Expression) -> Expression:
    """Retrieves the month for a datetime column.

    Returns:
        Expression: a UInt32 expression with just the month extracted from a datetime column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import month
        >>> df = daft.from_pydict(
        ...     {
        ...         "datetime": [
        ...             datetime.datetime(2024, 7, 3, 0, 0, 0),
        ...             datetime.datetime(2024, 6, 4, 0, 0, 0),
        ...             datetime.datetime(2024, 5, 5, 0, 0, 0),
        ...         ],
        ...     }
        ... )
        >>> df.with_column("month", month(df["datetime"])).collect()
        ╭─────────────────────┬────────╮
        │ datetime            ┆ month  │
        │ ---                 ┆ ---    │
        │ Timestamp[us]       ┆ UInt32 │
        ╞═════════════════════╪════════╡
        │ 2024-07-03 00:00:00 ┆ 7      │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ 2024-06-04 00:00:00 ┆ 6      │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌┤
        │ 2024-05-05 00:00:00 ┆ 5      │
        ╰─────────────────────┴────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("month", expr)


def quarter(expr: Expression) -> Expression:
    """Retrieves the quarter for a datetime column.

    Returns:
        Expression: a UInt32 expression with just the quarter extracted from a datetime column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import quarter
        >>> df = daft.from_pydict(
        ...     {
        ...         "datetime": [
        ...             datetime.datetime(2024, 1, 1, 0, 0, 0),
        ...             datetime.datetime(2023, 7, 4, 0, 0, 0),
        ...             datetime.datetime(2022, 12, 5, 0, 0, 0),
        ...         ],
        ...     }
        ... )
        >>> df.with_column("quarter", quarter(df["datetime"])).collect()
        ╭─────────────────────┬─────────╮
        │ datetime            ┆ quarter │
        │ ---                 ┆ ---     │
        │ Timestamp[us]       ┆ UInt32  │
        ╞═════════════════════╪═════════╡
        │ 2024-01-01 00:00:00 ┆ 1       │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
        │ 2023-07-04 00:00:00 ┆ 3       │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
        │ 2022-12-05 00:00:00 ┆ 4       │
        ╰─────────────────────┴─────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("quarter", expr)


def year(expr: Expression) -> Expression:
    """Retrieves the year for a datetime column.

    Returns:
        Expression: a Int32 expression with just the year extracted from a datetime column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import year
        >>> df = daft.from_pydict(
        ...     {
        ...         "datetime": [
        ...             datetime.datetime(2024, 7, 3, 0, 0, 0),
        ...             datetime.datetime(2023, 7, 4, 0, 0, 0),
        ...             datetime.datetime(2022, 7, 5, 0, 0, 0),
        ...         ],
        ...     }
        ... )
        >>> df.with_column("year", year(df["datetime"])).collect()
        ╭─────────────────────┬───────╮
        │ datetime            ┆ year  │
        │ ---                 ┆ ---   │
        │ Timestamp[us]       ┆ Int32 │
        ╞═════════════════════╪═══════╡
        │ 2024-07-03 00:00:00 ┆ 2024  │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 2023-07-04 00:00:00 ┆ 2023  │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 2022-07-05 00:00:00 ┆ 2022  │
        ╰─────────────────────┴───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("year", expr)


def day_of_week(expr: Expression) -> Expression:
    """Retrieves the day of the week for a datetime column, starting at 0 for Monday and ending at 6 for Sunday.

    Returns:
        Expression: a UInt32 expression with just the day_of_week extracted from a datetime column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import day_of_week
        >>> df = daft.from_pydict(
        ...     {
        ...         "datetime": [
        ...             datetime.datetime(2024, 7, 3, 0, 0, 0),
        ...             datetime.datetime(2024, 7, 4, 0, 0, 0),
        ...             datetime.datetime(2024, 7, 5, 0, 0, 0),
        ...         ],
        ...     }
        ... )
        >>> df.with_column("day_of_week", day_of_week(df["datetime"])).collect()
        ╭─────────────────────┬─────────────╮
        │ datetime            ┆ day_of_week │
        │ ---                 ┆ ---         │
        │ Timestamp[us]       ┆ UInt32      │
        ╞═════════════════════╪═════════════╡
        │ 2024-07-03 00:00:00 ┆ 2           │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2024-07-04 00:00:00 ┆ 3           │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2024-07-05 00:00:00 ┆ 4           │
        ╰─────────────────────┴─────────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("day_of_week", expr)


def day_of_month(expr: Expression) -> Expression:
    """Retrieves the day of the month for a datetime column.

    Returns:
        Expression: a UInt32 expression with just the day_of_month extracted from a datetime column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import day_of_month
        >>> df = daft.from_pydict(
        ...     {
        ...         "datetime": [
        ...             datetime.datetime(2024, 1, 1, 0, 0, 0),
        ...             datetime.datetime(2024, 2, 1, 0, 0, 0),
        ...             datetime.datetime(2024, 12, 31, 0, 0, 0),
        ...             datetime.datetime(2023, 12, 31, 0, 0, 0),
        ...         ],
        ...     }
        ... )
        >>> df.with_column("day_of_month", day_of_month(df["datetime"])).collect()
        ╭─────────────────────┬──────────────╮
        │ datetime            ┆ day_of_month │
        │ ---                 ┆ ---          │
        │ Timestamp[us]       ┆ UInt32       │
        ╞═════════════════════╪══════════════╡
        │ 2024-01-01 00:00:00 ┆ 1            │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2024-02-01 00:00:00 ┆ 1            │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2024-12-31 00:00:00 ┆ 31           │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2023-12-31 00:00:00 ┆ 31           │
        ╰─────────────────────┴──────────────╯
        <BLANKLINE>
        (Showing first 4 of 4 rows)
    """
    return Expression._call_builtin_scalar_fn("day_of_month", expr)


def day_of_year(expr: Expression) -> Expression:
    """Retrieves the ordinal day for a datetime column. Starting at 1 for January 1st and ending at 365 or 366 for December 31st.

    Returns:
        Expression: a UInt32 expression with just the day_of_year extracted from a datetime column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import day_of_year
        >>> df = daft.from_pydict(
        ...     {
        ...         "datetime": [
        ...             datetime.datetime(2024, 1, 1, 0, 0, 0),
        ...             datetime.datetime(2024, 2, 1, 0, 0, 0),
        ...             datetime.datetime(2024, 12, 31, 0, 0, 0),  # 2024 is a leap year
        ...             datetime.datetime(2023, 12, 31, 0, 0, 0),  # not leap year
        ...         ],
        ...     }
        ... )
        >>> df.with_column("day_of_year", day_of_year(df["datetime"])).collect()
        ╭─────────────────────┬─────────────╮
        │ datetime            ┆ day_of_year │
        │ ---                 ┆ ---         │
        │ Timestamp[us]       ┆ UInt32      │
        ╞═════════════════════╪═════════════╡
        │ 2024-01-01 00:00:00 ┆ 1           │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2024-02-01 00:00:00 ┆ 32          │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2024-12-31 00:00:00 ┆ 366         │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2023-12-31 00:00:00 ┆ 365         │
        ╰─────────────────────┴─────────────╯
        <BLANKLINE>
        (Showing first 4 of 4 rows)
    """
    return Expression._call_builtin_scalar_fn("day_of_year", expr)


def week_of_year(expr: Expression) -> Expression:
    """Retrieves the week of the year for a datetime column.

    Returns:
        Expression: a UInt32 expression with just the week_of_year extracted from a datetime column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import week_of_year
        >>> df = daft.from_pydict(
        ...     {
        ...         "datetime": [
        ...             datetime.datetime(2024, 1, 1, 0, 0, 0),
        ...             datetime.datetime(2024, 2, 1, 0, 0, 0),
        ...             datetime.datetime(
        ...                 2024, 12, 31, 0, 0, 0
        ...             ),  # part of week 1 of 2025 according to ISO 8601 standard
        ...             datetime.datetime(2023, 12, 31, 0, 0, 0),
        ...         ],
        ...     }
        ... )
        >>> df.with_column("week_of_year", week_of_year(df["datetime"])).collect()
        ╭─────────────────────┬──────────────╮
        │ datetime            ┆ week_of_year │
        │ ---                 ┆ ---          │
        │ Timestamp[us]       ┆ UInt32       │
        ╞═════════════════════╪══════════════╡
        │ 2024-01-01 00:00:00 ┆ 1            │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2024-02-01 00:00:00 ┆ 5            │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2024-12-31 00:00:00 ┆ 1            │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2023-12-31 00:00:00 ┆ 52           │
        ╰─────────────────────┴──────────────╯
        <BLANKLINE>
        (Showing first 4 of 4 rows)
    """
    return Expression._call_builtin_scalar_fn("week_of_year", expr)


def strftime(expr: Expression, format: str | None = None) -> Expression:
    """Converts a datetime/date column to a string column.

    Args:
        expr: The datetime or date expression to convert.
        format: The format to use for the conversion. If None, defaults to ISO 8601 format.

    Note:
        The format must be a valid datetime format string. (defaults to ISO 8601 format)
        See: https://docs.rs/chrono/latest/chrono/format/strftime/index.html


    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import strftime
        >>> df = daft.from_pydict(
        ...     {
        ...         "dates": [datetime.date(2023, 1, 1), datetime.date(2023, 1, 2), datetime.date(2023, 1, 3)],
        ...         "datetimes": [
        ...             datetime.datetime(2023, 1, 1, 12, 1),
        ...             datetime.datetime(2023, 1, 2, 12, 0, 0, 0),
        ...             datetime.datetime(2023, 1, 3, 12, 0, 0, 999_999),
        ...         ],
        ...     }
        ... )
        >>> df = df.with_column("datetimes_s", df["datetimes"].cast(daft.DataType.timestamp("s")))
        >>> df.select(
        ...     strftime(df["dates"]).alias("iso_date"),
        ...     strftime(df["dates"], format="%m/%d/%Y").alias("custom_date"),
        ...     strftime(df["datetimes"]).alias("iso_datetime"),
        ...     strftime(df["datetimes_s"]).alias("iso_datetime_s"),
        ...     strftime(df["datetimes_s"], format="%Y/%m/%d %H:%M:%S").alias("custom_datetime"),
        ... ).show()
        ╭────────────┬─────────────┬────────────────────────────┬─────────────────────┬─────────────────────╮
        │ iso_date   ┆ custom_date ┆ iso_datetime               ┆ iso_datetime_s      ┆ custom_datetime     │
        │ ---        ┆ ---         ┆ ---                        ┆ ---                 ┆ ---                 │
        │ String     ┆ String      ┆ String                     ┆ String              ┆ String              │
        ╞════════════╪═════════════╪════════════════════════════╪═════════════════════╪═════════════════════╡
        │ 2023-01-01 ┆ 01/01/2023  ┆ 2023-01-01T12:01:00        ┆ 2023-01-01T12:01:00 ┆ 2023/01/01 12:01:00 │
        ├╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2023-01-02 ┆ 01/02/2023  ┆ 2023-01-02T12:00:00        ┆ 2023-01-02T12:00:00 ┆ 2023/01/02 12:00:00 │
        ├╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2023-01-03 ┆ 01/03/2023  ┆ 2023-01-03T12:00:00.999999 ┆ 2023-01-03T12:00:00 ┆ 2023/01/03 12:00:00 │
        ╰────────────┴─────────────┴────────────────────────────┴─────────────────────┴─────────────────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    return Expression._call_builtin_scalar_fn("strftime", expr, format)


def total_seconds(expr: Expression) -> Expression:
    """Calculates the total number of seconds for a duration column.

    Returns:
        Expression: a UInt64 expression with the total number of seconds for a duration column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import total_seconds
        >>> df = daft.from_pydict(
        ...     {
        ...         "duration": [
        ...             datetime.timedelta(seconds=1),
        ...             datetime.timedelta(milliseconds=1),
        ...             datetime.timedelta(microseconds=1),
        ...             datetime.timedelta(days=1),
        ...             datetime.timedelta(hours=1),
        ...             datetime.timedelta(minutes=1),
        ...         ]
        ...     }
        ... )
        >>> df.with_column("Total Seconds", total_seconds(df["duration"])).show()
        ╭──────────────┬───────────────╮
        │ duration     ┆ Total Seconds │
        │ ---          ┆ ---           │
        │ Duration[us] ┆ Int64         │
        ╞══════════════╪═══════════════╡
        │ 1s           ┆ 1             │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1000µs       ┆ 0             │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1µs          ┆ 0             │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1d           ┆ 86400         │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1h           ┆ 3600          │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1m           ┆ 60            │
        ╰──────────────┴───────────────╯
        <BLANKLINE>
        (Showing first 6 of 6 rows)
    """
    return Expression._call_builtin_scalar_fn("total_seconds", expr)


def total_milliseconds(expr: Expression) -> Expression:
    """Calculates the total number of milliseconds for a duration column.

    Returns:
        Expression: a UInt64 expression with the total number of milliseconds for a duration column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import total_milliseconds
        >>> df = daft.from_pydict(
        ...     {
        ...         "duration": [
        ...             datetime.timedelta(seconds=1),
        ...             datetime.timedelta(milliseconds=1),
        ...             datetime.timedelta(microseconds=1),
        ...             datetime.timedelta(days=1),
        ...             datetime.timedelta(hours=1),
        ...             datetime.timedelta(minutes=1),
        ...         ]
        ...     }
        ... )
        >>> df.with_column("Total Milliseconds", total_milliseconds(df["duration"])).show()
        ╭──────────────┬────────────────────╮
        │ duration     ┆ Total Milliseconds │
        │ ---          ┆ ---                │
        │ Duration[us] ┆ Int64              │
        ╞══════════════╪════════════════════╡
        │ 1s           ┆ 1000               │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1000µs       ┆ 1                  │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1µs          ┆ 0                  │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1d           ┆ 86400000           │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1h           ┆ 3600000            │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1m           ┆ 60000              │
        ╰──────────────┴────────────────────╯
        <BLANKLINE>
        (Showing first 6 of 6 rows)
    """
    return Expression._call_builtin_scalar_fn("total_milliseconds", expr)


def total_microseconds(expr: Expression) -> Expression:
    """Calculates the total number of microseconds for a duration column.

    Returns:
        Expression: a UInt64 expression with the total number of microseconds for a duration column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import total_microseconds
        >>> df = daft.from_pydict(
        ...     {
        ...         "duration": [
        ...             datetime.timedelta(seconds=1),
        ...             datetime.timedelta(milliseconds=1),
        ...             datetime.timedelta(microseconds=1),
        ...             datetime.timedelta(days=1),
        ...             datetime.timedelta(hours=1),
        ...             datetime.timedelta(minutes=1),
        ...         ]
        ...     }
        ... )
        >>> df.with_column("Total Microseconds", total_microseconds(df["duration"])).show()
        ╭──────────────┬────────────────────╮
        │ duration     ┆ Total Microseconds │
        │ ---          ┆ ---                │
        │ Duration[us] ┆ Int64              │
        ╞══════════════╪════════════════════╡
        │ 1s           ┆ 1000000            │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1000µs       ┆ 1000               │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1µs          ┆ 1                  │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1d           ┆ 86400000000        │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1h           ┆ 3600000000         │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1m           ┆ 60000000           │
        ╰──────────────┴────────────────────╯
        <BLANKLINE>
        (Showing first 6 of 6 rows)
    """
    return Expression._call_builtin_scalar_fn("total_microseconds", expr)


def total_nanoseconds(expr: Expression) -> Expression:
    """Calculates the total number of nanoseconds for a duration column.

    Returns:
        Expression: a UInt64 expression with the total number of nanoseconds for a duration column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import total_nanoseconds
        >>> df = daft.from_pydict(
        ...     {
        ...         "duration": [
        ...             datetime.timedelta(seconds=1),
        ...             datetime.timedelta(milliseconds=1),
        ...             datetime.timedelta(microseconds=1),
        ...             datetime.timedelta(days=1),
        ...             datetime.timedelta(hours=1),
        ...             datetime.timedelta(minutes=1),
        ...         ]
        ...     }
        ... )
        >>> df.with_column("Total Nanoseconds", total_nanoseconds(df["duration"])).show()
        ╭──────────────┬───────────────────╮
        │ duration     ┆ Total Nanoseconds │
        │ ---          ┆ ---               │
        │ Duration[us] ┆ Int64             │
        ╞══════════════╪═══════════════════╡
        │ 1s           ┆ 1000000000        │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1000µs       ┆ 1000000           │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1µs          ┆ 1000              │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1d           ┆ 86400000000000    │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1h           ┆ 3600000000000     │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1m           ┆ 60000000000       │
        ╰──────────────┴───────────────────╯
        <BLANKLINE>
        (Showing first 6 of 6 rows)
    """
    return Expression._call_builtin_scalar_fn("total_nanoseconds", expr)


def total_minutes(expr: Expression) -> Expression:
    """Calculates the total number of minutes for a duration column.

    Returns:
        Expression: a UInt64 expression with the total number of minutes for a duration column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import total_minutes
        >>> df = daft.from_pydict(
        ...     {
        ...         "duration": [
        ...             datetime.timedelta(seconds=1),
        ...             datetime.timedelta(milliseconds=1),
        ...             datetime.timedelta(microseconds=1),
        ...             datetime.timedelta(days=1),
        ...             datetime.timedelta(hours=1),
        ...             datetime.timedelta(minutes=1),
        ...         ]
        ...     }
        ... )
        >>> df.with_column("Total Minutes", total_minutes(df["duration"])).show()
        ╭──────────────┬───────────────╮
        │ duration     ┆ Total Minutes │
        │ ---          ┆ ---           │
        │ Duration[us] ┆ Int64         │
        ╞══════════════╪═══════════════╡
        │ 1s           ┆ 0             │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1000µs       ┆ 0             │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1µs          ┆ 0             │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1d           ┆ 1440          │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1h           ┆ 60            │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1m           ┆ 1             │
        ╰──────────────┴───────────────╯
        <BLANKLINE>
        (Showing first 6 of 6 rows)
    """
    return Expression._call_builtin_scalar_fn("total_minutes", expr)


def total_hours(expr: Expression) -> Expression:
    """Calculates the total number of hours for a duration column.

    Returns:
        Expression: a UInt64 expression with the total number of hours for a duration column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import total_hours
        >>> df = daft.from_pydict(
        ...     {
        ...         "duration": [
        ...             datetime.timedelta(seconds=1),
        ...             datetime.timedelta(milliseconds=1),
        ...             datetime.timedelta(microseconds=1),
        ...             datetime.timedelta(days=1),
        ...             datetime.timedelta(hours=1),
        ...             datetime.timedelta(minutes=1),
        ...         ]
        ...     }
        ... )
        >>> df.with_column("Total Hours", total_hours(df["duration"])).show()
        ╭──────────────┬─────────────╮
        │ duration     ┆ Total Hours │
        │ ---          ┆ ---         │
        │ Duration[us] ┆ Int64       │
        ╞══════════════╪═════════════╡
        │ 1s           ┆ 0           │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1000µs       ┆ 0           │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1µs          ┆ 0           │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1d           ┆ 24          │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1h           ┆ 1           │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1m           ┆ 0           │
        ╰──────────────┴─────────────╯
        <BLANKLINE>
        (Showing first 6 of 6 rows)
    """
    return Expression._call_builtin_scalar_fn("total_hours", expr)


def total_days(expr: Expression) -> Expression:
    """Calculates the total number of days for a duration column.

    Returns:
        Expression: a UInt64 expression with the total number of days for a duration column

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import total_days
        >>> df = daft.from_pydict(
        ...     {
        ...         "duration": [
        ...             datetime.timedelta(seconds=1),
        ...             datetime.timedelta(milliseconds=1),
        ...             datetime.timedelta(microseconds=1),
        ...             datetime.timedelta(days=1),
        ...             datetime.timedelta(hours=1),
        ...             datetime.timedelta(minutes=1),
        ...         ]
        ...     }
        ... )
        >>> df.with_column("Total Days", total_days(df["duration"])).show()
        ╭──────────────┬────────────╮
        │ duration     ┆ Total Days │
        │ ---          ┆ ---        │
        │ Duration[us] ┆ Int64      │
        ╞══════════════╪════════════╡
        │ 1s           ┆ 0          │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1000µs       ┆ 0          │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1µs          ┆ 0          │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1d           ┆ 1          │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1h           ┆ 0          │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 1m           ┆ 0          │
        ╰──────────────┴────────────╯
        <BLANKLINE>
        (Showing first 6 of 6 rows)
    """
    return Expression._call_builtin_scalar_fn("total_days", expr)


def to_date(expr: Expression, format: str) -> Expression:
    """Converts a string to a date using the specified format.

    Returns:
        Expression: a Date expression which is parsed by given format

    Note:
        The format must be a valid date format string. See: https://docs.rs/chrono/latest/chrono/format/strftime/index.html

    Examples:
        >>> import daft
        >>> from daft.functions import to_date
        >>> df = daft.from_pydict({"x": ["2021-01-01", "2021-01-02", None]})
        >>> df = df.with_column("date", to_date(df["x"], "%Y-%m-%d"))
        >>> df.show()
        ╭────────────┬────────────╮
        │ x          ┆ date       │
        │ ---        ┆ ---        │
        │ String     ┆ Date       │
        ╞════════════╪════════════╡
        │ 2021-01-01 ┆ 2021-01-01 │
        ├╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2021-01-02 ┆ 2021-01-02 │
        ├╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ None       ┆ None       │
        ╰────────────┴────────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("to_date", expr, format=format)


def to_datetime(expr: Expression, format: str, timezone: str | None = None) -> Expression:
    """Converts a string to a datetime using the specified format and timezone.

    Returns:
        Expression: a DateTime expression which is parsed by given format and timezone

    Note:
        The format must be a valid datetime format string. See: https://docs.rs/chrono/latest/chrono/format/strftime/index.html

    Examples:
        >>> import daft
        >>> from daft.functions import to_datetime
        >>> df = daft.from_pydict({"x": ["2021-01-01 00:00:00.123", "2021-01-02 12:30:00.456", None]})
        >>> df = df.with_column("datetime", to_datetime(df["x"], "%Y-%m-%d %H:%M:%S%.3f"))
        >>> df.show()
        ╭─────────────────────────┬─────────────────────────╮
        │ x                       ┆ datetime                │
        │ ---                     ┆ ---                     │
        │ String                  ┆ Timestamp[ms]           │
        ╞═════════════════════════╪═════════════════════════╡
        │ 2021-01-01 00:00:00.123 ┆ 2021-01-01 00:00:00.123 │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2021-01-02 12:30:00.456 ┆ 2021-01-02 12:30:00.456 │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ None                    ┆ None                    │
        ╰─────────────────────────┴─────────────────────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

        If a timezone is provided, the datetime will be parsed in that timezone

        >>> df = daft.from_pydict({"x": ["2021-01-01 00:00:00.123 +0800", "2021-01-02 12:30:00.456 +0800", None]})
        >>> df = df.with_column("datetime", to_datetime(df["x"], "%Y-%m-%d %H:%M:%S%.3f %z", timezone="Asia/Shanghai"))
        >>> df.show()
        ╭───────────────────────────────┬──────────────────────────────╮
        │ x                             ┆ datetime                     │
        │ ---                           ┆ ---                          │
        │ String                        ┆ Timestamp[ms; Asia/Shanghai] │
        ╞═══════════════════════════════╪══════════════════════════════╡
        │ 2021-01-01 00:00:00.123 +0800 ┆ 2021-01-01 00:00:00.123 CST  │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2021-01-02 12:30:00.456 +0800 ┆ 2021-01-02 12:30:00.456 CST  │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ None                          ┆ None                         │
        ╰───────────────────────────────┴──────────────────────────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("to_datetime", expr, format=format, timezone=timezone)


def date_trunc(interval: str, expr: Expression, relative_to: Expression | None = None) -> Expression:
    """Truncates the datetime column to the specified interval.

    Args:
        interval: The interval to truncate to. Must be a string representing a valid interval in "{integer} {unit}" format, e.g. "1 day". Valid time units are: 'microsecond', 'millisecond', 'second', 'minute', 'hour', 'day', 'week'.
        expr: The datetime expression to truncate.
        relative_to (optional): Timestamp to truncate relative to. If not provided, truncates to the start of the Unix epoch: 1970-01-01 00:00:00.

    Returns:
        Expression: a DateTime expression truncated to the specified interval

    Examples:
        >>> import datetime
        >>> import daft
        >>> from daft.functions import date_trunc
        >>>
        >>> df = daft.from_pydict(
        ...     {
        ...         "datetime": [
        ...             datetime.datetime(2021, 1, 1, 0, 1, 1),
        ...             datetime.datetime(2021, 1, 1, 0, 1, 59),
        ...             datetime.datetime(2021, 1, 1, 0, 2, 0),
        ...         ],
        ...     }
        ... )
        >>> df.with_column("truncated", date_trunc("1 minute", df["datetime"])).collect()
        ╭─────────────────────┬─────────────────────╮
        │ datetime            ┆ truncated           │
        │ ---                 ┆ ---                 │
        │ Timestamp[us]       ┆ Timestamp[us]       │
        ╞═════════════════════╪═════════════════════╡
        │ 2021-01-01 00:01:01 ┆ 2021-01-01 00:01:00 │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2021-01-01 00:01:59 ┆ 2021-01-01 00:01:00 │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2021-01-01 00:02:00 ┆ 2021-01-01 00:02:00 │
        ╰─────────────────────┴─────────────────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

    """
    return Expression._call_builtin_scalar_fn("truncate", expr, relative_to, interval=interval)


def to_unix_epoch(expr: Expression, time_unit: str | TimeUnit | None = None) -> Expression:
    """Converts a datetime column to a Unix timestamp with the specified time unit. (default: seconds).

    See [daft.datatype.TimeUnit](https://docs.daft.ai/en/stable/api/datatypes/all_datatypes/#daft.datatype.DataType.timeunit) for more information on time units and valid values.

    Examples:
        >>> import daft
        >>> from datetime import date
        >>>
        >>> df = daft.from_pydict(
        ...     {
        ...         "dates": [
        ...             date(2001, 1, 1),
        ...             date(2001, 1, 2),
        ...             date(2001, 1, 3),
        ...             None,
        ...         ]
        ...     }
        ... )
        >>> df.with_column("timestamp", df["dates"].to_unix_epoch("ns")).show()
        ╭────────────┬────────────────────╮
        │ dates      ┆ timestamp          │
        │ ---        ┆ ---                │
        │ Date       ┆ Int64              │
        ╞════════════╪════════════════════╡
        │ 2001-01-01 ┆ 978307200000000000 │
        ├╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2001-01-02 ┆ 978393600000000000 │
        ├╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ 2001-01-03 ┆ 978480000000000000 │
        ├╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ None       ┆ None               │
        ╰────────────┴────────────────────╯
        <BLANKLINE>
        (Showing first 4 of 4 rows)
    """
    return Expression._call_builtin_scalar_fn("to_unix_epoch", expr, time_unit=time_unit)



"""Audio Functions."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import daft
from daft.udf.udf_v2 import Func

if TYPE_CHECKING:
    from daft import Expression
    from daft.file.typing import AudioMetadata


def get_metadata_impl(
    file: daft.AudioFile,
) -> AudioMetadata:
    return file.metadata()


audio_metadata_fn = Func._from_func(
    get_metadata_impl,
    return_dtype=daft.DataType.struct(
        {
            "sample_rate": daft.DataType.int64(),
            "channels": daft.DataType.int64(),
            "frames": daft.DataType.float64(),
            "format": daft.DataType.string(),
            "subtype": daft.DataType.string(),
        }
    ),
    unnest=False,
    use_process=None,
    is_batch=False,
    batch_size=None,
    max_retries=None,
    on_error=None,
)


def audio_metadata(
    file_expr: Expression,
) -> Expression:
    """Get metadata for a audio file.

    Args:
        file_expr (AudioFile Expression): The audio file to get metadata for.

    Returns:
        Expression (Struct Expression): A struct containing the metadata
        of the audio file.

        The struct contains the following fields:
            - sample_rate: int - The sample rate of the audio file
            - channels: int - The number of channels in the audio file
            - frames: int - The number of frames in the audio file
            - format: str - The format of the audio file
            - subtype: str | None - The subtype of the audio file

    Examples:
        >>> import daft
        >>> from daft.functions import audio_file, audio_metadata, resample

        >>> # Discover the audio files (returns a dataframe with ['path', 'size'] columns)
        >>> df = daft.from_glob_path("hf://datasets/Eventual-Inc/sample-files/audio/*.mp3")

        >>> df = (
        ...     df.with_column("file", audio_file(daft.col("path")))
        ...     .with_column("metadata", audio_metadata(daft.col("file")))
        ...     .select("path", "size", "metadata")
        ... )
        >>> df.show(3)  # doctest: +SKIP
        ╭────────────────────────────────┬─────────┬──────────────────────────────────────────────────────────────────────╮
        │ path                           ┆ size    ┆ metadata                                                             │
        │ ---                            ┆ ---     ┆ ---                                                                  │
        │ String                         ┆ Int64   ┆ Struct[sample_rate: Int64, channels: Int64, frames: Float64, format: │
        │                                ┆         ┆ String, subtype: String]                                             │
        ╞════════════════════════════════╪═════════╪══════════════════════════════════════════════════════════════════════╡
        │ hf://datasets/Eventual-Inc/sa… ┆ 822924  ┆ {sample_rate: 16000,                                                 │
        │                                ┆         ┆ channels…                                                            │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ hf://datasets/Eventual-Inc/sa… ┆ 618408  ┆ {sample_rate: 16000,                                                 │
        │                                ┆         ┆ channels…                                                            │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ hf://datasets/Eventual-Inc/sa… ┆ 1190736 ┆ {sample_rate: 16000,                                                 │
        │                                ┆         ┆ channels…                                                            │
        ╰────────────────────────────────┴─────────┴──────────────────────────────────────────────────────────────────────╯
        <BLANKLINE>
        (Showing first 3 rows)
    """
    # Check for required dependencies at expression definition time
    from daft.dependencies import sf

    if not sf.module_available():
        raise ImportError(
            "The 'soundfile' module is required to get audio metadata. "
            "Please install it with: pip install 'daft[audio]'"
        )

    return audio_metadata_fn(file_expr)


def resample_impl(
    file: daft.AudioFile,
    sample_rate: int,
) -> Any:
    return file.resample(sample_rate)


resample_fn = Func._from_func(
    resample_impl,
    return_dtype=daft.DataType.tensor(daft.DataType.float64()),
    unnest=False,
    use_process=None,
    is_batch=False,
    batch_size=None,
    max_retries=None,
    on_error=None,
)


def resample(
    file_expr: Expression,
    sample_rate: int,
) -> Expression:
    """Resample a audio file.

    Args:
        file_expr (AudioFile Expression): The audio file to resample.
        sample_rate (int): The sample rate to resample to.

    Returns:
        Expression (Tensor[Python] Expression): The resampled audio file.

    Example:
        >>> import daft
        >>> from daft.functions import audio_file, resample

        >>> # Discover the audio files (returns a dataframe with ['path', 'size'] columns)
        >>> df = daft.from_glob_path("hf://datasets/Eventual-Inc/sample-files/audio/*.mp3")

        >>> df = (
        ...     df.with_column("file", audio_file(daft.col("path")))
        ...     .with_column("resampled", resample(daft.col("file"), sample_rate=16000))
        ...     .select("path", "size", "resampled")
        ... )
        >>> df.show(3)
        ╭────────────────────────────────┬─────────┬──────────────────────────╮
        │ path                           ┆ size    ┆ resampled                │
        │ ---                            ┆ ---     ┆ ---                      │
        │ String                         ┆ Int64   ┆ Tensor[Float64]          │
        ╞════════════════════════════════╪═════════╪══════════════════════════╡
        │ hf://datasets/Eventual-Inc/sa… ┆ 822924  ┆ <Tensor shape=(2490278)> │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ hf://datasets/Eventual-Inc/sa… ┆ 618408  ┆ <Tensor shape=(2207923)> │
        ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
        │ hf://datasets/Eventual-Inc/sa… ┆ 1190736 ┆ <Tensor shape=(3341800)> │
        ╰────────────────────────────────┴─────────┴──────────────────────────╯
        <BLANKLINE>
        (Showing first 3 rows)
    """
    # Check for required dependencies at expression definition time
    # This gives users a helpful error immediately, rather than during execution
    from daft.dependencies import librosa, sf

    if not sf.module_available():
        raise ImportError(
            "The 'soundfile' module is required to resample audio files. "
            "Please install it with: pip install 'daft[audio]'"
        )
    if not librosa.module_available():
        raise ImportError(
            "The 'librosa' module is required to resample audio files. "
            "Please install it with: pip install 'daft[audio]'"
        )

    return resample_fn(file_expr, sample_rate)



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from daft.scarf_telemetry import track_import_on_scarf

###
# Set up code coverage for when running code coverage with ray
###
if "COV_CORE_SOURCE" in os.environ:
    try:
        from pytest_cov.embed import init

        init()
    except Exception as exc:
        import sys

        sys.stderr.write(
            "pytest-cov: Failed to setup subprocess coverage. Environ: {!r} Exception: {!r}\n".format(
                {k: v for k, v in os.environ.items() if k.startswith("COV_CORE")}, exc
            )
        )

###
# Get build constants from Rust .so
###

from daft.daft import build_type as _build_type
from daft.daft import version as _version
from daft.daft import refresh_logger as _refresh_logger


def get_version() -> str:
    return _version()


def get_build_type() -> str:
    return _build_type()


def refresh_logger() -> None:
    """Refreshes Daft's internal rust logging to the current python log level."""
    _refresh_logger()


__version__ = get_version()

###
# Initialize analytics
###

track_import_on_scarf()

###
# Daft top-level imports
###

from daft.catalog import (
    Catalog,
    Identifier,
    Table,
)
from daft.context import (
    set_execution_config,
    set_planning_config,
    execution_config_ctx,
    planning_config_ctx,
)
from daft.convert import (
    from_arrow,
    from_dask_dataframe,
    from_pandas,
    from_pydict,
    from_pylist,
    from_ray_dataset,
)
from daft.daft import ImageFormat, ImageMode, ImageProperty, ResourceRequest
from daft.dataframe import DataFrame
from daft.schema import Schema
from daft.datatype import DataType, TimeUnit, MediaType
from daft.expressions import Expression, col, element, lit, interval
from daft.series import Series
from daft.session import (
    Session,
    attach,
    attach_catalog,
    attach_provider,
    attach_function,
    attach_table,
    create_namespace,
    create_namespace_if_not_exists,
    create_table,
    create_table_if_not_exists,
    create_temp_table,
    current_catalog,
    current_model,
    current_namespace,
    current_provider,
    current_session,
    detach_catalog,
    detach_function,
    detach_provider,
    detach_table,
    drop_namespace,
    drop_table,
    get_catalog,
    get_function,
    get_provider,
    get_table,
    has_catalog,
    has_namespace,
    has_provider,
    has_table,
    list_catalogs,
    list_tables,
    load_extension,
    read_table,
    session,
    set_catalog,
    set_model,
    set_namespace,
    set_provider,
    set_session,
    write_table,
)
from daft.udf import udf, func, cls, method, metrics
from daft.io import (
    DataCatalogTable,
    DataCatalogType,
    IOConfig,
    from_glob_path,
    _range as range,
    read_csv,
    read_deltalake,
    read_hudi,
    read_iceberg,
    read_json,
    read_parquet,
    read_sql,
    read_text,
    read_video_frames,
    read_warc,
    read_huggingface,
    read_mcap,
)
from daft.runners import get_or_create_runner, get_or_infer_runner_type, set_runner_native, set_runner_ray
from daft.sql import sql, sql_expr
from daft.viz import register_viz_hook
from daft.window import Window
from daft.file import File, VideoFile, AudioFile

from daft import context
from daft import io
from daft import runners
from daft import datasets
from daft import functions
from daft import gravitino


# Lance is lazy-loaded because lance_namespace pulls in ~450ms of pydantic models.
if TYPE_CHECKING:
    from daft.io import read_lance


def __getattr__(name: str) -> object:
    if name == "read_lance":
        return getattr(io, name)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = [
    "AudioFile",
    "Catalog",
    "DataCatalogTable",
    "DataCatalogType",
    "DataFrame",
    "DataType",
    "Expression",
    "File",
    "IOConfig",
    "Identifier",
    "ImageFormat",
    "ImageMode",
    "ImageProperty",
    "MediaType",
    "ResourceRequest",
    "Schema",
    "Series",
    "Session",
    "Table",
    "TimeUnit",
    "VideoFile",
    "Window",
    "attach",
    "attach_catalog",
    "attach_function",
    "attach_provider",
    "attach_table",
    "cls",
    "col",
    "context",
    "create_namespace",
    "create_namespace_if_not_exists",
    "create_table",
    "create_table_if_not_exists",
    "create_temp_table",
    "current_catalog",
    "current_model",
    "current_namespace",
    "current_provider",
    "current_session",
    "datasets",
    "detach_catalog",
    "detach_function",
    "detach_provider",
    "detach_table",
    "drop_namespace",
    "drop_table",
    "element",
    "execution_config_ctx",
    "from_arrow",
    "from_dask_dataframe",
    "from_glob_path",
    "from_pandas",
    "from_pydict",
    "from_pylist",
    "from_ray_dataset",
    "func",
    "functions",
    "get_catalog",
    "get_function",
    "get_or_create_runner",
    "get_or_infer_runner_type",
    "get_provider",
    "get_table",
    "gravitino",
    "has_catalog",
    "has_namespace",
    "has_provider",
    "has_table",
    "interval",
    "io",
    "list_catalogs",
    "list_tables",
    "lit",
    "load_extension",
    "method",
    "metrics",
    "planning_config_ctx",
    "range",
    "read_csv",
    "read_deltalake",
    "read_hudi",
    "read_huggingface",
    "read_iceberg",
    "read_json",
    "read_lance",
    "read_mcap",
    "read_parquet",
    "read_sql",
    "read_table",
    "read_text",
    "read_video_frames",
    "read_warc",
    "refresh_logger",
    "register_viz_hook",
    "runners",
    "session",
    "set_catalog",
    "set_execution_config",
    "set_model",
    "set_namespace",
    "set_planning_config",
    "set_provider",
    "set_runner_native",
    "set_runner_ray",
    "set_session",
    "sql",
    "sql_expr",
    "udf",
    "write_table",
]



from __future__ import annotations

import asyncio
import uuid
from typing import TYPE_CHECKING, Any

from daft.expressions.expressions import Expression, ExpressionsProjection
from daft.recordbatch.micropartition import MicroPartition
from daft.runners.ray_compat import validate_and_normalize_ray_options

if TYPE_CHECKING:
    from ray.actor import ActorHandle as RayActorHandle

    from daft.daft import PyExpr, PyMicroPartition

try:
    import ray
except ImportError:
    raise

MAX_UDFACTOR_ACTOR_RESTARTS = 4
MAX_UDFACTOR_ACTOR_TASK_RETRIES = 4


@ray.remote(
    max_restarts=MAX_UDFACTOR_ACTOR_RESTARTS,
    max_task_retries=MAX_UDFACTOR_ACTOR_TASK_RETRIES,
)
class UDFActor:
    def __init__(self, uninitialized_projection: ExpressionsProjection) -> None:
        self.projection = ExpressionsProjection([e._initialize_udfs() for e in uninitialized_projection])

    def get_node_id(self) -> str:
        return ray.get_runtime_context().get_node_id()

    def eval_input(self, input: PyMicroPartition) -> PyMicroPartition:
        mp = MicroPartition._from_pymicropartition(input)
        res = mp.eval_expression_list(self.projection)
        return res._micropartition


class UDFActorHandle:
    def __init__(self, actor_ref: RayActorHandle) -> None:
        self.actor = actor_ref

    def actor_id(self) -> str:
        return self.actor._actor_id.hex()

    async def eval_input(self, input: PyMicroPartition) -> PyMicroPartition:
        return await self.actor.eval_input.remote(input)

    def teardown(self) -> None:
        ray.kill(self.actor)


def get_ready_actors_by_location(
    actor_handles: list[UDFActorHandle],
) -> tuple[list[UDFActorHandle], list[UDFActorHandle]]:
    from ray._private.state import actors

    current_node_id = ray.get_runtime_context().get_node_id()

    local_actors = []
    remote_actors = []
    for actor_handle in actor_handles:
        actor_id = actor_handle.actor_id()
        actor_state = actors(actor_id)
        if actor_state["Address"]["NodeID"] == current_node_id:
            local_actors.append(actor_handle)
        else:
            remote_actors.append(actor_handle)

    return local_actors, remote_actors


async def start_udf_actors(
    projection: PyExpr,
    num_actors: int,
    num_gpus_per_actor: float,
    num_cpus_per_actor: float,
    memory_per_actor: float,
    ray_options: dict[str, Any] | None,
    timeout: int,
    actor_name: str | None = None,
) -> list[UDFActorHandle]:
    expr_projection = ExpressionsProjection([Expression._from_pyexpr(projection)])

    udf_options = validate_and_normalize_ray_options(
        {
            "scheduling_strategy": "SPREAD",
            "num_gpus": num_gpus_per_actor,
            "num_cpus": num_cpus_per_actor,
            "memory": memory_per_actor,
            **(ray_options or {}),
        }
    )

    actors: list[RayActorHandle] = [
        UDFActor.options(  # type: ignore
            name=None if actor_name is None else f"{actor_name}:{str(uuid.uuid4())[:8]}-{rank}",
            **udf_options,
        ).remote(expr_projection)
        for rank in range(num_actors)
    ]

    # Wait for actors to be ready
    ready_futures = [asyncio.wrap_future(actor.__ray_ready__.remote().future()) for actor in actors]
    ready_refs, _ = await asyncio.wait(ready_futures, return_when=asyncio.ALL_COMPLETED, timeout=timeout)

    # Verify that the __ray_ready__ calls were successful
    await asyncio.gather(*ready_refs)

    if not ready_refs:
        raise RuntimeError(
            f"UDF actors failed to start within {timeout} seconds, please increase the actor_udf_ready_timeout config "
            f"via daft.set_execution_config(actor_udf_ready_timeout=timeout)"
        )

    # Return the ready actors
    ready_indices = [ready_futures.index(ref) for ref in ready_refs]
    ready_actors = [UDFActorHandle(actors[i]) for i in ready_indices]
    return ready_actors



"""Human-readable query name generator.

Produces Docker-style random names like ``eager-phoenix-a4f821`` for use as
query identifiers.  With 55 adjectives x 55 nouns x 16^6 hex suffixes the
combinatorial space is ~48 billion, which is more than sufficient for
global uniqueness across tracing systems.
"""

from __future__ import annotations

import os

_ADJECTIVES = (
    "agile",
    "bold",
    "brave",
    "bright",
    "calm",
    "clever",
    "cool",
    "cosmic",
    "daring",
    "eager",
    "epic",
    "fair",
    "fancy",
    "fast",
    "fierce",
    "fleet",
    "fresh",
    "gentle",
    "glad",
    "golden",
    "grand",
    "happy",
    "hardy",
    "keen",
    "kind",
    "lively",
    "lucky",
    "merry",
    "mighty",
    "neat",
    "nimble",
    "noble",
    "plucky",
    "polite",
    "proud",
    "quick",
    "quiet",
    "rapid",
    "sharp",
    "shiny",
    "sleek",
    "smart",
    "snappy",
    "solar",
    "steady",
    "stellar",
    "strong",
    "sunny",
    "super",
    "swift",
    "tidy",
    "tough",
    "vivid",
    "warm",
    "witty",
)

_NOUNS = (
    "aurora",
    "badger",
    "bear",
    "comet",
    "condor",
    "coral",
    "crane",
    "dolphin",
    "eagle",
    "falcon",
    "finch",
    "flame",
    "forest",
    "fox",
    "galaxy",
    "grove",
    "hawk",
    "heron",
    "island",
    "jade",
    "lake",
    "lark",
    "lily",
    "lotus",
    "lynx",
    "maple",
    "meadow",
    "nebula",
    "oak",
    "ocean",
    "otter",
    "owl",
    "panda",
    "pearl",
    "phoenix",
    "pine",
    "plaza",
    "quasar",
    "raven",
    "reef",
    "river",
    "sage",
    "sierra",
    "sparrow",
    "star",
    "stone",
    "summit",
    "tiger",
    "valley",
    "wave",
    "willow",
    "wind",
    "wolf",
    "wren",
    "zenith",
)


def generate_query_name() -> str:
    """Return a human-readable query name like ``eager-phoenix-a4f821``."""
    raw = os.urandom(5)
    adj_idx = raw[0] % len(_ADJECTIVES)
    noun_idx = raw[1] % len(_NOUNS)
    hex_suffix = raw[2:5].hex()
    return f"{_ADJECTIVES[adj_idx]}-{_NOUNS[noun_idx]}-{hex_suffix}"



from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from openai import AsyncOpenAI, OpenAIError, RateLimitError
from openai import OpenAI as OpenAIClient
from openai._types import Omit, omit
from openai.types.create_embedding_response import Usage

from daft import DataType
from daft.ai.metrics import record_token_metrics
from daft.ai.openai.typing import OpenAIProviderOptions
from daft.ai.protocols import TextEmbedder, TextEmbedderDescriptor
from daft.ai.typing import EmbeddingDimensions, EmbedTextOptions, Options, UDFOptions
from daft.ai.utils import merge_provider_and_api_options
from daft.dependencies import np

if TYPE_CHECKING:
    from openai.types import EmbeddingModel
    from openai.types.create_embedding_response import CreateEmbeddingResponse

    from daft.ai.typing import Embedding


@dataclass(frozen=True)
class _ModelProfile:
    """Model profiles contain various model-specific metadata.

    Note:
        This is a bit simpler than OO-inheritance to model the subtle
        differences between the models. If there is a need for different
        implementations, then it would make sense to
    """

    dimensions: EmbeddingDimensions
    supports_overriding_dimensions: bool
    input_text_token_limit: int = 8192  # Default per-input token limit


_models: dict[EmbeddingModel, _ModelProfile] = {
    "text-embedding-ada-002": _ModelProfile(
        dimensions=EmbeddingDimensions(
            size=1536,
            dtype=DataType.float32(),
        ),
        supports_overriding_dimensions=False,
    ),
    "text-embedding-3-small": _ModelProfile(
        dimensions=EmbeddingDimensions(
            size=1536,
            dtype=DataType.float32(),
        ),
        supports_overriding_dimensions=True,
    ),
    "text-embedding-3-large": _ModelProfile(
        dimensions=EmbeddingDimensions(
            size=3072,
            dtype=DataType.float32(),
        ),
        supports_overriding_dimensions=True,
    ),
}


def get_input_text_token_limit_for_model(model_name: str) -> int:
    """Get the input token limit for a model, with fallback to default.

    Args:
        model_name: The name of the embedding model.

    Returns:
        The input text token limit for the model. Returns 8192 for unknown/custom models.
    """
    if model_name in _models:
        return _models[model_name].input_text_token_limit
    else:
        return 8192  # Default for unknown/custom models


@dataclass
class OpenAITextEmbedderDescriptor(TextEmbedderDescriptor):
    provider_name: str
    provider_options: OpenAIProviderOptions
    model_name: str
    dimensions: int | None
    embed_options: EmbedTextOptions = field(
        default_factory=lambda: EmbedTextOptions(batch_size=64, max_retries=3, on_error="raise")
    )

    def __post_init__(self) -> None:
        if self.provider_options.get("base_url") is None:
            if self.model_name not in _models:
                supported_models = ", ".join(_models.keys())
                raise ValueError(
                    f"Unsupported OpenAI embedding model '{self.model_name}', expected one of: {supported_models}"
                )
            model = _models[self.model_name]
            if self.dimensions is not None:
                if model.supports_overriding_dimensions:
                    if "supports_overriding_dimensions" not in self.embed_options:
                        self.embed_options["supports_overriding_dimensions"] = True
                else:
                    raise ValueError(
                        f"OpenAI embedding model '{self.model_name}' does not support specifying dimensions"
                    )

    def get_provider(self) -> str:
        return self.provider_name

    def get_model(self) -> str:
        return self.model_name

    def get_options(self) -> Options:
        return dict(self.embed_options)

    def get_dimensions(self) -> EmbeddingDimensions:
        if self.dimensions is not None:
            return EmbeddingDimensions(size=self.dimensions, dtype=_models[self.model_name].dimensions.dtype)

        if self.provider_options.get("base_url") is not None and self.model_name not in _models:
            try:
                merged_provider_options: dict[str, Any] = merge_provider_and_api_options(
                    provider_options=self.provider_options,
                    api_options=self.embed_options,
                    provider_option_type=OpenAIProviderOptions,
                )

                client = OpenAIClient(**merged_provider_options)
                response = client.embeddings.create(
                    input="dimension probe",
                    model=self.model_name,
                    encoding_format="float",
                )
                size = len(response.data[0].embedding)
                return EmbeddingDimensions(size=size, dtype=DataType.float32())
            except Exception as ex:
                raise ValueError(
                    "Failed to determine embedding dimensions from OpenAI-compatible embedding server. "
                    "Specify `dimensions=...` or ensure the server supports embeddings.create."
                ) from ex
        else:
            return _models[self.model_name].dimensions

    def get_udf_options(self) -> UDFOptions:
        options = super().get_udf_options()
        options.max_retries = 0  # OpenAI client handles retries internally
        return options

    def is_async(self) -> bool:
        return True

    def instantiate(self) -> TextEmbedder:
        # Get batch_token_limit from embed_options, default to 300_000
        batch_token_limit = self.embed_options.get("batch_token_limit", 300_000)

        # Get input_text_token_limit from model profile using helper function
        input_text_token_limit = get_input_text_token_limit_for_model(self.model_name)

        return OpenAITextEmbedder(
            provider_options=self.provider_options,
            model=self.model_name,
            embed_options=self.embed_options,
            dimensions=self.dimensions if self.embed_options.get("supports_overriding_dimensions", False) else omit,
            provider_name=self.provider_name,
            batch_token_limit=batch_token_limit,
            input_text_token_limit=input_text_token_limit,
        )


class OpenAITextEmbedder(TextEmbedder):
    """The OpenAI TextEmbedder will batch across rows, and split a large row into a batch request when necessary.

    Note:
        Token limits are configurable via batch_token_limit (provider-level, defaults to 300k)
        and input_text_token_limit (model-level, defaults to 8,192).
        This implementation uses len(text) // 3 to estimate token count,
        which is conservative and O(1) rather than being perfect with tiktoken.
    """

    _client: AsyncOpenAI
    _model: str
    _dimensions: int | None
    _batch_token_limit: int
    _input_text_token_limit: int

    def __init__(
        self,
        provider_options: OpenAIProviderOptions,
        model: str,
        embed_options: EmbedTextOptions,
        dimensions: int | None = None,
        zero_on_failure: bool = False,
        provider_name: str = "openai",
        batch_token_limit: int = 300_000,
        input_text_token_limit: int = 8192,
    ):
        self._model = model
        self._zero_on_failure = zero_on_failure
        self._dimensions = dimensions
        self._provider_name = provider_name
        self._batch_token_limit = batch_token_limit
        self._input_text_token_limit = input_text_token_limit

        merged_provider_options: dict[str, Any] = merge_provider_and_api_options(
            provider_options=provider_options,
            api_options=embed_options,
            provider_option_type=OpenAIProviderOptions,
        )

        self._client = AsyncOpenAI(**merged_provider_options)

    async def embed_text(self, text: list[str]) -> list[Embedding]:
        embeddings: list[Embedding] = []
        curr_batch: list[str] = []
        curr_batch_token_count: int = 0

        approx_chars_per_token = 3  # round down for conservative estimate of "1 token ≈ 4 characters"
        input_text_chars_limit = self._input_text_token_limit * approx_chars_per_token

        async def flush() -> None:
            nonlocal curr_batch
            nonlocal curr_batch_token_count
            if len(curr_batch) == 0:
                return None
            embeddings_result = await self._embed_text_batch(curr_batch)
            embeddings.extend(embeddings_result)
            curr_batch = []
            curr_batch_token_count = 0

        for input_text in text:
            # Handle None values by treating them as empty strings
            if input_text is None:
                input_text = ""
            input_text_token_count = len(input_text) // approx_chars_per_token
            if input_text_token_count > self._input_text_token_limit:
                # Must process previous inputs first, if any, to maintain ordered outputs.
                await flush()
                # If the current input exceeds the maximum tokens per input (configurable, default 8192),
                # then we will split this single input into its own batch request.
                chunked_batch = chunk_text(input_text, input_text_chars_limit)
                chunked_result = await self._embed_text_batch(chunked_batch)
                # We combine all result embedding vectors into a single embedding using a weighted average.
                # https://github.com/openai/openai-cookbook/blob/main/examples/Embedding_long_inputs.ipynb
                chunked_lens = [len(chunk) for chunk in chunked_batch]
                chunked_vec = np.average(chunked_result, axis=0, weights=chunked_lens)
                chunked_vec = chunked_vec / np.linalg.norm(chunked_vec)  # normalizes length to 1
                embeddings.append(chunked_vec)
            elif input_text_token_count + curr_batch_token_count >= self._batch_token_limit:
                await flush()
                curr_batch.append(input_text)
                curr_batch_token_count += input_text_token_count
            else:
                curr_batch.append(input_text)
                curr_batch_token_count += input_text_token_count
        await flush()

        return embeddings

    async def _embed_text_batch(self, input_batch: list[str]) -> list[Embedding]:
        """Embeds text as a batch call, falling back to _embed_text on rate limit exceptions."""
        try:
            dimensions: int | Omit = self._dimensions if self._dimensions is not None else omit
            response = await self._client.embeddings.create(
                input=input_batch,
                model=self._model,
                encoding_format="float",
                dimensions=dimensions,
            )
            self._record_usage_metrics(response)
            return [np.array(embedding.embedding) for embedding in response.data]
        except RateLimitError:
            # fall back to individual calls when rate limited
            # consider sleeping or other backoff mechanisms
            return await asyncio.gather(*(self._embed_text(text) for text in input_batch))
        except OpenAIError as ex:
            raise ValueError("The `embed_text` method encountered an OpenAI error.") from ex

    async def _embed_text(self, input_text: str) -> Embedding:
        """Embeds a single text input and possibly returns a zero vector."""
        try:
            dimensions: int | Omit = self._dimensions if self._dimensions is not None else omit
            response: CreateEmbeddingResponse = await self._client.embeddings.create(
                input=input_text,
                model=self._model,
                encoding_format="float",
                dimensions=dimensions,
            )
            self._record_usage_metrics(response)
            return np.array(response.data[0].embedding)
        except Exception as ex:
            if self._zero_on_failure:
                model_profile = _models[self._model]
                dtype = model_profile.dimensions.dtype
                size = self._dimensions or model_profile.dimensions.size
                np_dtype = np.float64 if dtype == DataType.float64() else np.float32
                return np.zeros(size, dtype=np_dtype)
            else:
                raise ex

    def _record_usage_metrics(self, response: CreateEmbeddingResponse) -> None:
        usage = getattr(response, "usage", None)
        if usage is None or not isinstance(usage, Usage):
            return

        input_tokens = usage.prompt_tokens
        total_tokens = usage.total_tokens

        record_token_metrics(
            protocol="embed",
            model=self._model,
            provider=self._provider_name,
            input_tokens=input_tokens,
            total_tokens=total_tokens,
        )


def chunk_text(text: str, size: int) -> list[str]:
    return [text[i : i + size] for i in range(0, len(text), size)]



from __future__ import annotations

import logging
import sys
from typing import TYPE_CHECKING

from daft.context import get_context
from daft.daft import (
    FileFormatConfig,
    FileInfos,
    IOConfig,
    LocalPhysicalPlan,
    PyQueryResult,
    QueryEndState,
    set_compute_runtime_num_worker_threads,
)
from daft.errors import UDFException
from daft.execution.metadata import ExecutionMetadata
from daft.execution.native_executor import NativeExecutor
from daft.filesystem import glob_path_with_stats
from daft.naming import generate_query_name
from daft.recordbatch import MicroPartition
from daft.runners import runner_io
from daft.runners.partitioning import (
    LocalMaterializedResult,
    LocalPartitionSet,
    PartitionCacheEntry,
    PartitionSetCache,
)
from daft.runners.runner import LOCAL_PARTITION_SET_CACHE, Runner
from daft.scarf_telemetry import track_runner_on_scarf

if TYPE_CHECKING:
    from collections.abc import Generator, Iterator

    from daft.logical.builder import LogicalPlanBuilder

logger = logging.getLogger(__name__)


class NativeRunnerIO(runner_io.RunnerIO):
    def glob_paths_details(
        self,
        source_paths: list[str],
        file_format_config: FileFormatConfig | None = None,
        io_config: IOConfig | None = None,
    ) -> FileInfos:
        file_infos = FileInfos()
        file_format = file_format_config.file_format() if file_format_config is not None else None
        for source_path in set(source_paths):
            try:
                path_file_infos = glob_path_with_stats(source_path, file_format, io_config)
                file_infos.merge(path_file_infos)
            except FileNotFoundError:
                logger.debug("%s is not found.", source_path)

        if len(file_infos) == 0:
            raise FileNotFoundError(f"No files found at {','.join(source_paths)}")

        return file_infos


class NativeRunner(Runner[MicroPartition]):
    name = "native"

    def __init__(self, num_threads: int | None = None) -> None:
        super().__init__()
        if num_threads is not None:
            set_compute_runtime_num_worker_threads(num_threads)

        self.native_executor = NativeExecutor()

    def initialize_partition_set_cache(self) -> PartitionSetCache:
        return LOCAL_PARTITION_SET_CACHE

    def runner_io(self) -> NativeRunnerIO:
        return NativeRunnerIO()

    def run_iter(
        self,
        builder: LogicalPlanBuilder,
        results_buffer_size: int | None = None,
    ) -> Generator[LocalMaterializedResult, None, ExecutionMetadata]:
        track_runner_on_scarf(runner=self.name)

        # NOTE: Freeze and use this same execution config for the entire execution
        ctx = get_context()
        query_id = generate_query_name()
        output_schema = builder.schema()

        entrypoint = "python " + " ".join(sys.argv)

        try:
            # Try to send notifications, but don't fail the query if they fail
            from daft.daft import PyQueryMetadata

            ctx._notify_query_start(
                query_id,
                PyQueryMetadata(
                    output_schema._schema,
                    builder.repr_json(),
                    "Native (Swordfish)",
                    ray_dashboard_url=None,
                    entrypoint=entrypoint,
                ),
            )
            ctx._notify_optimization_start(query_id)
        except Exception as e:
            logger.warning("Failed to send notifications: %s", e)
            pass

        # Optimize the logical plan.
        builder = builder.optimize(ctx.daft_execution_config)

        try:
            ctx._notify_optimization_end(query_id, builder.repr_json())
        except Exception as e:
            logger.warning("Failed to send optimization end notification: %s", e)
            pass

        psets = {
            k: [v.micropartition()._micropartition for v in v.values()]
            for k, v in self._part_set_cache.get_all_partition_sets().items()
        }
        plan, inputs = LocalPhysicalPlan.from_logical_plan_builder(builder._builder, psets)

        results_gen = self.native_executor.run(
            plan,
            inputs,
            ctx,
            {"query_id": query_id},
        )

        try:
            while True:
                result = next(results_gen)
                try:
                    ctx._notify_result_out(query_id, result.partition())
                except Exception:
                    pass
                yield result
        except GeneratorExit:
            # Generator was abandoned (e.g., .show() breaking out early after collecting
            # enough rows). Notify subscribers so the dashboard transitions out of Finalizing.
            try:
                query_result = PyQueryResult(QueryEndState.Finished, "Query finished")
                ctx._notify_query_end(query_id, query_result)
            except Exception as e:
                logger.warning("Failed to send query end notification: %s", e)
            return  # type: ignore[return-value]
        except StopIteration as e:
            query_result = PyQueryResult(QueryEndState.Finished, "Query finished")
            ctx._notify_query_end(query_id, query_result)
            physical_plan_json, execution_stats = e.value
            return ExecutionMetadata._from_runner_output(execution_stats, query_id, physical_plan_json)
        except KeyboardInterrupt as e:
            query_result = PyQueryResult(QueryEndState.Canceled, "Query canceled by the user.")
            ctx._notify_query_end(query_id, query_result)
            raise e
        except UDFException as e:
            err_msg = f"UDF failed with exception: {e.original_exception}"
            query_result = PyQueryResult(QueryEndState.Failed, err_msg)
            ctx._notify_query_end(query_id, query_result)
            raise e
        except Exception as e:
            err_msg = f"General Exception raised: {e}"
            query_result = PyQueryResult(QueryEndState.Failed, err_msg)
            ctx._notify_query_end(query_id, query_result)
            raise e

    def run_iter_tables(
        self, builder: LogicalPlanBuilder, results_buffer_size: int | None = None
    ) -> Iterator[MicroPartition]:
        for result in self.run_iter(builder, results_buffer_size=results_buffer_size):
            yield result.partition()

    def run(self, builder: LogicalPlanBuilder) -> tuple[PartitionCacheEntry, ExecutionMetadata]:
        results_gen = self.run_iter(builder)
        result_pset = LocalPartitionSet()

        try:
            i = 0
            while True:
                result = next(results_gen)
                result_pset.set_partition(i, result)
                i += 1
        except StopIteration as e:
            metadata = e.value

        pset_entry = self.put_partition_set_into_cache(result_pset)
        return pset_entry, metadata



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Borrowed and modified from [`skypilot`](https://github.com/skypilot-org/skypilot/blob/master/sky/adaptors/common.py).
from __future__ import annotations

import importlib
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from types import ModuleType


class LazyImport:
    """Manages Optional Dependency imports.

    There are certain large imports (e.g. Ray, daft.unity_catalog.UnityCatalogTable, etc.) that
    do not need to be top-level imports. For example, Ray should only be imported when the ray
    runner is used, or specific ray data extension types are needed. We can lazily import these
    modules as needed.
    """

    def __init__(self, module_name: str):
        self._module_name = module_name
        self._module: ModuleType | None = None

    def module_available(self) -> bool:
        try:
            self._load_module()
            return True
        except ImportError:
            return False

    def _load_module(self) -> ModuleType:
        if self._module is None:
            try:
                self._module = importlib.import_module(self._module_name)
            except ImportError as e:
                raise ImportError(f"Failed to import lazily-loaded module '{self._module_name}'.") from e
        return self._module

    def __getattr__(self, name: str) -> Any:
        # Given a lazy module and an attribute to get, we have the following possibilities:
        #   1. The attribute is the lazy object's attribute.
        #   2. The attribute is an attribute of the module.
        #   3. The module does not exist.
        #   4. The attribute is a submodule.
        #   5. The attribute does not exist.
        try:
            if name in self.__dict__:
                return self.__dict__[name]
            return getattr(self._load_module(), name)
        except AttributeError as e:
            # Dynamically create a new LazyImport instance for the submodule.
            submodule_name = f"{self._module_name}.{name}"
            lazy_submodule = LazyImport(submodule_name)
            if lazy_submodule.module_available():
                setattr(self, name, lazy_submodule)
                return lazy_submodule
            raise e

    def __getstate__(self) -> dict[str, Any]:
        # Only serialize the module name, don't serialize the loaded module to avoid pickling issues
        return {"_module_name": self._module_name}

    def __setstate__(self, state: dict[str, Any]) -> None:
        # Restore the module name and initialize module as None
        self._module_name = state["_module_name"]
        self._module = None



from __future__ import annotations

import logging
import warnings
from typing import TYPE_CHECKING

from pyiceberg.io.pyarrow import schema_to_pyarrow
from pyiceberg.schema import Schema as IcebergSchema
from pyiceberg.schema import visit

import daft
from daft.daft import (
    CountMode,
    FileFormatConfig,
    ParquetSourceConfig,
    PyPartitionField,
    PyPartitionTransform,
    PyPushdowns,
    PyRecordBatch,
    ScanTask,
    StorageConfig,
)
from daft.datatype import DataType
from daft.dependencies import pa
from daft.io.iceberg.schema_field_id_mapping_visitor import SchemaFieldIdMappingVisitor
from daft.io.scan import ScanOperator, make_partition_field
from daft.logical.schema import Field, Schema
from daft.recordbatch import RecordBatch

if TYPE_CHECKING:
    from collections.abc import Iterator

    from pyiceberg.partitioning import PartitionField as IcebergPartitionField
    from pyiceberg.partitioning import PartitionSpec as IcebergPartitionSpec
    from pyiceberg.table import Table
    from pyiceberg.typedef import Record

logger = logging.getLogger(__name__)


def _iceberg_count_result_function(total_count: int, field_name: str) -> Iterator[PyRecordBatch]:
    """Construct Iceberg count query result."""
    try:
        arrow_schema = pa.schema([pa.field(field_name, pa.uint64())])
        arrow_array = pa.array([total_count], type=pa.uint64())
        arrow_batch = pa.RecordBatch.from_arrays([arrow_array], [field_name])

        logger.debug("Generated Iceberg count result: %s=%d", field_name, total_count)

        yield RecordBatch.from_arrow_record_batches([arrow_batch], arrow_schema)._recordbatch
    except Exception as e:
        logger.error("Failed to construct Iceberg count result: %s", e)
        raise


def _iceberg_partition_field_to_daft_partition_field(
    iceberg_schema: IcebergSchema, pfield: IcebergPartitionField
) -> PyPartitionField:
    name = pfield.name
    source_id = pfield.source_id
    source_field = iceberg_schema.find_field(source_id)
    source_name = source_field.name
    daft_field = Field.create(
        source_name, DataType.from_arrow_type(schema_to_pyarrow(iceberg_schema.find_type(source_name)))
    )
    transform = pfield.transform
    source_type = DataType.from_arrow_type(schema_to_pyarrow(source_field.field_type))

    from pyiceberg.transforms import (
        BucketTransform,
        DayTransform,
        HourTransform,
        IdentityTransform,
        MonthTransform,
        TruncateTransform,
        YearTransform,
    )

    tfm = None
    if isinstance(transform, IdentityTransform):
        tfm = PyPartitionTransform.identity()
        result_type = source_type
    elif isinstance(transform, YearTransform):
        tfm = PyPartitionTransform.year()
        result_type = DataType.int32()
    elif isinstance(transform, MonthTransform):
        tfm = PyPartitionTransform.month()
        result_type = DataType.int32()
    elif isinstance(transform, DayTransform):
        tfm = PyPartitionTransform.day()
        # pyiceberg uses date as the result type of a day transform, which is incorrect
        # so we cannot use transform.result_type() here
        result_type = DataType.date()
    elif isinstance(transform, HourTransform):
        tfm = PyPartitionTransform.hour()
        result_type = DataType.int32()
    elif isinstance(transform, BucketTransform):
        n = transform.num_buckets
        tfm = PyPartitionTransform.iceberg_bucket(n)
        result_type = DataType.int32()
    elif isinstance(transform, TruncateTransform):
        w = transform.width
        tfm = PyPartitionTransform.iceberg_truncate(w)
        result_type = source_type
    else:
        warnings.warn(f"{transform} not implemented, Please make an issue!")
        result_type = source_type
    result_field = Field.create(name, result_type)
    return make_partition_field(result_field, daft_field, transform=tfm)


def iceberg_partition_spec_to_fields(
    iceberg_schema: IcebergSchema, spec: IcebergPartitionSpec
) -> list[PyPartitionField]:
    return [_iceberg_partition_field_to_daft_partition_field(iceberg_schema, field) for field in spec.fields]


class IcebergScanOperator(ScanOperator):
    def __init__(self, iceberg_table: Table, snapshot_id: int | None, storage_config: StorageConfig) -> None:
        super().__init__()
        self._table = iceberg_table
        self._snapshot_id = snapshot_id
        self._storage_config = storage_config

        iceberg_schema = (
            iceberg_table.schema()
            if self._snapshot_id is None
            else self._table.scan(snapshot_id=self._snapshot_id).projection()
        )
        arrow_schema = schema_to_pyarrow(iceberg_schema)
        self._field_id_mapping = visit(iceberg_schema, SchemaFieldIdMappingVisitor())
        self._schema = Schema.from_pyarrow_schema(arrow_schema)

        self._partition_keys = iceberg_partition_spec_to_fields(iceberg_schema, self._table.spec())

    def schema(self) -> Schema:
        return self._schema

    def name(self) -> str:
        return "IcebergScanOperator"

    def display_name(self) -> str:
        return f"IcebergScanOperator({'.'.join(self._table.name())})"

    def partitioning_keys(self) -> list[PyPartitionField]:
        return self._partition_keys

    def _iceberg_record_to_partition_spec(
        self, spec: IcebergPartitionSpec, record: Record
    ) -> daft.recordbatch.RecordBatch | None:
        partition_fields = iceberg_partition_spec_to_fields(self._table.schema(), spec)
        arrays = dict()
        assert len(record) == len(partition_fields)
        for idx, pfield in enumerate(partition_fields):
            field = Field._from_pyfield(pfield.field)
            field_name = field.name
            field_dtype = field.dtype
            arrow_type = field_dtype.to_arrow_dtype()
            arrays[field_name] = daft.Series.from_arrow(pa.array([record[idx]], type=arrow_type), name=field_name).cast(
                field_dtype
            )
        if len(arrays) > 0:
            return daft.recordbatch.RecordBatch.from_pydict(arrays)
        else:
            return None

    def multiline_display(self) -> list[str]:
        return [
            self.display_name(),
            f"Schema = {self._schema}",
            f"Partitioning keys = {self.partitioning_keys}",
            # TODO(Clark): Improve repr of storage config here.
            f"Storage config = {self._storage_config}",
        ]

    def to_scan_tasks(self, pushdowns: PyPushdowns) -> Iterator[ScanTask]:
        # Check if there is a count aggregation pushdown
        if (
            pushdowns.aggregation is not None
            and pushdowns.aggregation_count_mode() is not None
            and pushdowns.aggregation_required_column_names()
        ):
            count_mode = pushdowns.aggregation_count_mode()
            fields = pushdowns.aggregation_required_column_names()

            if count_mode in self.supported_count_modes():
                logger.info(
                    "Using Iceberg count pushdown optimization for count mode: %s",
                    count_mode,
                )
                yield from self._create_count_scan_task(pushdowns, fields[0])
                return
            else:
                logger.warning(
                    "Count mode %s is not supported for pushdown, falling back to regular scan",
                    count_mode,
                )

        # Regular scan without count pushdown
        yield from self._create_regular_scan_tasks(pushdowns)

    def _create_regular_scan_tasks(self, pushdowns: PyPushdowns) -> Iterator[ScanTask]:
        """Create regular scan tasks without count pushdown."""
        limit = pushdowns.limit
        iceberg_tasks = self._table.scan(limit=limit, snapshot_id=self._snapshot_id).plan_files()

        limit_files = limit is not None and pushdowns.filters is None and pushdowns.partition_filters is None

        if len(self.partitioning_keys()) > 0 and pushdowns.partition_filters is None:
            logger.warning(
                "%s has Partitioning Keys: %s but no partition filter was specified. This will result in a full table scan.",
                self.display_name(),
                self.partitioning_keys(),
            )
        scan_tasks = []

        if limit is not None:
            rows_left = limit
        else:
            rows_left = 0
        for task in iceberg_tasks:
            if limit_files and (rows_left <= 0):
                break
            file = task.file
            path = file.file_path
            record_count = file.record_count
            file_format = file.file_format
            if file_format == "PARQUET":
                file_format_config = FileFormatConfig.from_parquet_config(
                    ParquetSourceConfig(field_id_mapping=self._field_id_mapping)
                )
            else:
                # TODO: Support ORC and AVRO when we can read it
                raise NotImplementedError(f"{file_format} for iceberg not implemented!")

            iceberg_delete_files = [f.file_path for f in task.delete_files]

            # TODO: Thread in Statistics to each ScanTask: P2
            pspec = self._iceberg_record_to_partition_spec(self._table.specs()[file.spec_id], file.partition)
            st = ScanTask.catalog_scan_task(
                file=path,
                file_format=file_format_config,
                schema=self._schema._schema,
                num_rows=record_count,
                storage_config=self._storage_config,
                size_bytes=file.file_size_in_bytes,
                iceberg_delete_files=iceberg_delete_files,
                pushdowns=pushdowns,
                partition_values=pspec._recordbatch if pspec is not None else None,
                stats=None,
            )
            if st is None:
                continue
            rows_left -= record_count
            scan_tasks.append(st)
        return iter(scan_tasks)

    def _create_count_scan_task(self, pushdowns: PyPushdowns, field_name: str) -> Iterator[ScanTask]:
        """Create count pushdown scan task using Iceberg metadata."""
        try:
            iceberg_tasks = self._table.scan(limit=None, snapshot_id=self._snapshot_id).plan_files()
            total_count = 0

            # Aggregate row counts from all data files
            for task in iceberg_tasks:
                data_file = task.file
                total_count += data_file.record_count

            result_schema = Schema.from_pyarrow_schema(pa.schema([pa.field(field_name, pa.uint64())]))

            logger.info("Created Iceberg count pushdown task with total_count=%d for field=%s", total_count, field_name)
            yield ScanTask.python_factory_func_scan_task(
                module=_iceberg_count_result_function.__module__,
                func_name=_iceberg_count_result_function.__name__,
                func_args=(total_count, field_name),
                schema=result_schema._schema,
                num_rows=1,
                size_bytes=8,
                pushdowns=pushdowns,
                stats=None,
                source_name=self.display_name(),
            )
        except Exception as e:
            logger.error("Failed to create Iceberg count pushdown task: %s, now falling back to regular scan", e)
            yield from self._create_regular_scan_tasks(pushdowns)

    def _has_delete_files(self) -> bool:
        """Check if the table has any delete files.

        This method quickly scans the table to determine if there are any delete files
        present. If delete files are found, count pushdown should be disabled to avoid
        complex delete file processing logic.

        Returns:
            True if the table has delete files, False otherwise
        """
        try:
            # Get a limited scan to check for delete files
            iceberg_tasks = self._table.scan(
                limit=1,  # Only need to check if any delete files exist
                snapshot_id=self._snapshot_id,
            ).plan_files()

            # Check if any task has delete files
            for task in iceberg_tasks:
                if task.delete_files and len(task.delete_files) > 0:
                    logger.debug("Found delete files in table, count pushdown will be disabled")
                    return True
            return False

        except Exception as e:
            logger.warning("Error checking for delete files: %s, disabling count pushdown as precaution", e)
            return True

    def can_absorb_filter(self) -> bool:
        return False

    def can_absorb_limit(self) -> bool:
        return False

    def can_absorb_select(self) -> bool:
        return True

    def supports_count_pushdown(self) -> bool:
        return not self._has_delete_files()

    def supported_count_modes(self) -> list[CountMode]:
        return [CountMode.All]



from __future__ import annotations

import logging
import typing

from daft import refresh_logger

VALID_LEVELS: set[str] = {
    "DEBUG",
    "INFO",
    "WARNING",
    "WARN",
    "ERROR",
}

LOG_LEVEL_MAP = {
    "ERROR": logging.ERROR,
    "WARNING": logging.WARNING,
    "WARN": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
}


def setup_debug_logger() -> None:
    import warnings

    warnings.warn(
        "setup_debug_logger() is deprecated. Use setup_logger('debug') instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return setup_logger("debug")


def setup_logger(
    level: str = "debug",
    exclude_prefix: typing.Iterable[str] | None = None,
    daft_only: bool = False,
    logformat: str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    datefmt: str = "%Y-%m-%d %H:%M:%S.%s".format(),
) -> None:
    """Setup Daft logger with a specific log level, optional prefix filtering, and Rust sync.

    Args:
        level (str, optional): The log level to use. Valid options are `DEBUG`, `INFO`, `WARNING`, `ERROR`. Defaults to `DEBUG`.
        exclude_prefix (typing.Iterable[str] | None, optional): A list of prefixes to exclude from logging. Defaults to None.
        daft_only (bool, optional): Whether to only log messages from the Daft module. Defaults to False.
        logformat (str, optional): The log format to use. Defaults to "%(asctime)s - %(name)s - %(levelname)s - %(message)s".
        datefmt (str, optional): The date format to use. Defaults to '%Y-%m-%d %H:%M:%S.%s'.format().

    Raises:
        ValueError: If the log level is not valid.
    """
    if not level or level.upper() not in VALID_LEVELS:
        raise ValueError(f"Invalid log level '{level}'. Valid options: {VALID_LEVELS}")
    level = level.upper()

    if level == "WARN":
        level = "WARNING"

    logging.basicConfig(level=level, format=logformat, datefmt=datefmt)

    root_logger: logging.Logger = logging.getLogger()

    numeric_level: int = LOG_LEVEL_MAP[level]
    root_logger.setLevel(numeric_level)

    for handler in root_logger.handlers:
        handler.filters.clear()

    if daft_only:
        for handler in root_logger.handlers:
            handler.addFilter(lambda record: record.name.startswith("daft"))

    if exclude_prefix:
        for prefix in exclude_prefix:
            for handler in root_logger.handlers:
                handler.addFilter(lambda r, p=prefix: not r.name.startswith(p))  # type: ignore

    refresh_logger()



from __future__ import annotations

import shutil
import tempfile
from typing import TYPE_CHECKING

from daft.daft import PyDaftFile, PyFileReference
from daft.datatype import MediaType
from daft.dependencies import av, sf

if TYPE_CHECKING:
    from tempfile import _TemporaryFileWrapper

    from daft.file.audio import AudioFile
    from daft.file.video import VideoFile
    from daft.io import IOConfig


class File:
    """A file-like object for working with file contents in Daft.

    This is an abstract base class that provides a standard file interface compatible
    with Python's file protocol.

    The File object can be used with most Python libraries that accept file-like objects,
    and implements the standard read/seek/tell interface. Files are read-only in the
    current implementation.

    Examples:
        >>> import daft
        >>> from daft.functions import file
        >>> df = daft.from_pydict({"paths": ["data.json"]})
        >>> df = df.select(file(df["paths"]))
        >>>
        >>> @daft.func
        >>> def read_json(file: daft.File) -> str:
        >>>     import json
        >>>     with file.open() as f:
        >>>         data = json.load(f)
        >>>         return data["text"]
    """

    _inner: PyFileReference

    @staticmethod
    def _from_file_reference(reference: PyFileReference) -> File:
        instance = File.__new__(File)
        instance._inner = reference
        return instance

    def __init__(
        self, url: str, io_config: IOConfig | None = None, media_type: MediaType = MediaType.unknown()
    ) -> None:
        self._inner = PyFileReference._from_tuple((media_type._media_type, url, io_config))  # type: ignore

    def open(self) -> PyDaftFile:
        return PyDaftFile._from_file_reference(self._inner)

    def __str__(self) -> str:
        return self._inner.__str__()

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return False

    def seekable(self) -> bool:
        return True

    def isatty(self) -> bool:
        return False

    @property
    def path(self) -> str:
        """The full path or URL of the file.

        Returns:
            str: The file path or URL.

        Example:
            >>> import daft
            >>> f = daft.File("s3://bucket/path/to/data.csv")
            >>> f.path
            's3://bucket/path/to/data.csv'
        """
        return self._inner.path()

    @property
    def name(self) -> str:
        """The filename (basename) extracted from the file path or URL.

        Returns:
            str: The filename without directory components.

        Example:
            >>> import daft
            >>> f = daft.File("s3://bucket/path/to/data.csv")
            >>> f.name
            'data.csv'
        """
        return self._inner.name()

    def size(self) -> int:
        return PyDaftFile._from_file_reference(self._inner).size()

    def mime_type(self) -> str:
        """Attempts to determine the MIME type of the file.

        If the MIME type is undetectable, returns 'application/octet-stream'.
        """
        with self.open() as f:
            maybe_mime_type = f.guess_mime_type()
            return maybe_mime_type if maybe_mime_type else "application/octet-stream"

    def to_tempfile(self) -> _TemporaryFileWrapper[bytes]:
        """Create a temporary file with the contents of this file.

        Returns:
            _TemporaryFileWrapper[bytes]: The temporary file object.

        The temporary file will be automatically deleted when the returned context manager is closed.

        It's important to note that `to_tempfile` closes the original file object, so it CANNOT be used after calling this method.
        """
        with self.open() as f:
            temp_file = tempfile.NamedTemporaryFile(
                prefix="daft_",
            )
            f.seek(0)

            size = f.size()
            # if its either a really small file, or doesn't support range requests. Just read it normally
            if not f._supports_range_requests() or size < 1024:
                temp_file.write(f.read())
            else:
                shutil.copyfileobj(f, temp_file, length=size)
            # close it as `to_tempfile` is a consuming method
            f.close()
            temp_file.seek(0)

            return temp_file

    def is_video(self) -> bool:
        mimetype = self.mime_type()
        if mimetype.startswith("video/"):
            return True
        return False

    def is_audio(self) -> bool:
        mimetype = self.mime_type()
        if mimetype.startswith("audio/"):
            return True
        return False

    def as_video(self) -> VideoFile:
        """Convert to VideoFile if this file contains video data."""
        if not av.module_available():
            raise ImportError("The 'av' module is required to convert files to video.")
        # this is purposely inside the function, and after the `av` check
        # because using VideoFile means that the user has `av` installed
        from daft.file.video import VideoFile

        if not self.is_video():
            raise ValueError(f"File {self} is not a video file")

        cls = VideoFile.__new__(VideoFile)
        cls._inner = self._inner

        return cls

    def as_audio(self) -> AudioFile:
        """Convert to AudioFile if this file contains audio data."""
        if not sf.module_available():
            raise ImportError(
                "The 'soundfile' module is required to convert files to audio. "
                "Please install it with: pip install 'daft[audio]'"
            )
        # this is purposely inside the function, and after the `sf` check
        # because using AudioFile means that the user has `sf` installed
        from daft.file.audio import AudioFile

        if not self.is_audio():
            raise ValueError(f"File {self} is not an audio file")

        cls = AudioFile.__new__(AudioFile)
        cls._inner = self._inner

        return cls



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any

from daft.datatype import DataType
from daft.expressions import col, lit
from daft.functions import length, serialize
from daft.io import DataSink
from daft.io.sink import WriteResult
from daft.recordbatch import MicroPartition
from daft.schema import Schema

if TYPE_CHECKING:
    from collections.abc import Iterator
    from types import ModuleType

    from google.cloud.bigtable.table import Table

    from daft.dataframe import DataFrame


class BigtableDataSink(DataSink[dict[str, Any]]):
    """WriteSink for writing data to Google Cloud Bigtable."""

    @staticmethod
    def _import_bigtable() -> ModuleType:
        try:
            from google.cloud import bigtable

            return bigtable
        except ImportError:
            raise ImportError(
                "google-cloud-bigtable is not installed. Please install it using `pip install google-cloud-bigtable`"
            )

    @staticmethod
    def _import_bigtable_batcher() -> ModuleType:
        try:
            from google.cloud.bigtable import batcher

            return batcher
        except ImportError:
            raise ImportError(
                "google-cloud-bigtable is not installed. Please install it using `pip install google-cloud-bigtable`"
            )

    def __init__(
        self,
        project_id: str,
        instance_id: str,
        table_id: str,
        row_key_column: str,
        column_family_mappings: dict[str, str],
        client_kwargs: dict[str, Any] | None = None,
        write_kwargs: dict[str, Any] | None = None,
    ) -> None:
        self._project_id = project_id
        self._instance_id = instance_id
        self._table_id = table_id
        self._column_family_mappings = column_family_mappings
        self._row_key_column = row_key_column
        self._client_kwargs = client_kwargs or {}
        self._write_kwargs = write_kwargs or {}

        self._result_schema = Schema._from_field_name_and_types([("write_responses", DataType.python())])

    @staticmethod
    def _is_bigtable_compatible_type(dtype: DataType) -> bool:
        """Check if a DataType is compatible with Bigtable."""
        return dtype.is_integer() or dtype.is_binary() or dtype.is_string()

    @staticmethod
    def _validate_bigtable_parameters(
        schema: Schema, row_key_column: str, column_family_mappings: dict[str, str]
    ) -> None:
        """Validate Bigtable parameters against the DataFrame schema."""
        column_names = schema.column_names()
        # Validate that the row key column exists in the schema.
        if row_key_column not in column_names:
            raise ValueError(f'Row key column "{row_key_column}" not found in schema')

        # Validate that column family mappings cover all columns except the row key column.
        data_columns = [col for col in column_names if col != row_key_column]
        missing_columns = [col for col in data_columns if col not in column_family_mappings]
        if missing_columns:
            raise ValueError(
                f"Column family mappings missing for columns: {missing_columns}. "
                f"All columns except the row key column ({row_key_column}) must have a column family mapping."
            )

    def _preprocess_dataframe(self, df: DataFrame, serialize_incompatible_types: bool = True) -> DataFrame:
        # Validate Bigtable parameters against the DataFrame schema.
        self._validate_bigtable_parameters(df.schema(), self._row_key_column, self._column_family_mappings)

        df_to_write = df

        # Check for incompatible types and handle them based on whether serialize_incompatible_types is True.
        incompatible_columns = []

        for field in df_to_write.schema():
            if self._is_bigtable_compatible_type(field.dtype):
                continue
            incompatible_columns.append(field.name)

        if incompatible_columns:
            if serialize_incompatible_types:
                warnings.warn(
                    f"Bigtable only supports integer, binary, and string types. Found incompatible columns: "
                    f"{', '.join([f'{column_name}' for column_name in incompatible_columns])}. "
                    "These will be automatically serialized to JSON. Set `serialize_incompatible_types=False` to disable automatic conversion."
                )

                for col_name in incompatible_columns:
                    df_to_write = df_to_write.with_column(col_name, serialize(df_to_write[col_name], format="json"))
            else:
                raise ValueError(
                    f"Bigtable only supports integer, binary, and string types. Found incompatible columns: "
                    f"{', '.join([f'{column_name}' for column_name in incompatible_columns])}. "
                    "Set `serialize_incompatible_types=True` to automatically convert these to JSON, or convert them manually before writing."
                )

        # Filter out rows with invalid row keys.
        df_to_write = df_to_write.filter(
            col(self._row_key_column).not_null() & (length(col(self._row_key_column)) > lit(0))
        )

        return df_to_write

    def name(self) -> str:
        return "Bigtable Data Sink"

    def schema(self) -> Schema:
        return self._result_schema

    def _write_with_error_handling(self, table: Table, mp: MicroPartition) -> WriteResult[dict[str, Any]]:
        try:
            bigtable_batcher = BigtableDataSink._import_bigtable_batcher()
            rows_written = 0

            with bigtable_batcher.MutationsBatcher(table=table, **self._write_kwargs) as batcher:
                rows = []
                data_list = mp.to_pylist()

                for row_data in data_list:
                    row_key = str(row_data[self._row_key_column])
                    direct_row = table.direct_row(row_key)
                    for column_name, value in row_data.items():
                        if column_name != self._row_key_column and value is not None:
                            direct_row.set_cell(
                                self._column_family_mappings[column_name],
                                column_name,
                                value,
                            )

                    rows.append(direct_row)

                if rows:
                    batcher.mutate_rows(rows)
                    rows_written = len(rows)

            result = {
                "status": "success",
                "rows_written": rows_written,
            }
            return WriteResult(
                result=result,
                bytes_written=mp.size_bytes() or 0,
                rows_written=rows_written,
            )
        except Exception as e:
            result = {
                "status": "failed",
                "error": str(e),
                "rows_not_written": len(mp),
            }
            return WriteResult(
                result=result,
                bytes_written=0,
                rows_written=0,
            )

    def write(self, micropartitions: Iterator[MicroPartition]) -> Iterator[WriteResult[dict[str, Any]]]:
        bigtable = BigtableDataSink._import_bigtable()

        # Handle emulator host configuration.
        client_kwargs = self._client_kwargs.copy()
        if "emulator_host" in client_kwargs:
            import os

            os.environ["BIGTABLE_EMULATOR_HOST"] = client_kwargs.pop("emulator_host")

        client = bigtable.Client(project=self._project_id, admin=True, **client_kwargs)
        instance = client.instance(self._instance_id)

        for micropartition in micropartitions:
            if len(micropartition) == 0:
                continue

            table = instance.table(self._table_id)
            yield self._write_with_error_handling(table, micropartition)

    def finalize(self, write_results: list[WriteResult[dict[str, Any]]]) -> MicroPartition:
        if len(write_results) == 0:
            return MicroPartition.empty(self._result_schema)
        else:
            result_table = MicroPartition.from_pydict(
                {
                    "write_responses": write_results,
                }
            )

            return result_table
